# -*- coding: utf-8 -*-
"""3D volume discretizations

.. currentmodule:: torchgdm.struct.volume

Classes
-------

.. autosummary::
   :toctree: generated/
   
   StructDiscretized3D
   StructDiscretizedCubic3D
   StructDiscretizedHexagonal3D


Functions
---------

.. autosummary::
   :toctree: generated/
   
   extract_eff_pola_via_mp_decomposition
   extract_eff_pola_via_propagation


Geometries
----------

.. autosummary::
   :toctree: generated/
   
   cube
   cuboid
   sphere
   spheroid
   disc
   ellipse
   split_ring
   prism_trigonal
   from_image


Discretizer functions
---------------------

.. autosummary::
   :toctree: generated/
   
   discretizer_cubic
   discretizer_hexagonalcompact


"""
from .pola import StructDiscretizedHexagonal3D, StructDiscretizedCubic3D
from .pola import StructDiscretized3D
from .pola import extract_eff_pola_via_mp_decomposition
from .pola import extract_eff_pola_via_propagation
from . import pola
from . import geometries

from .geometries import discretizer_cubic
from .geometries import discretizer_hexagonalcompact

from .geometries import cube
from .geometries import cuboid
from .geometries import sphere
from .geometries import spheroid
from .geometries import disc
from .geometries import ellipse
from .geometries import split_ring
from .geometries import prism_trigonal

from .geometries import from_image



# encoding=utf-8
# %%
import unittest

import torch

import torchgdm as tg


class TestFarfieldVsCrosssect(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing far-field scattering vs scattering cross-section...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # --- setup a test case
        # - environment
        eps_env = 1.0
        mat_env = tg.materials.MatConstant(eps_env)
        env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=mat_env)

        # - illumination field(s)
        self.wavelengths = torch.tensor([532.0])
        e_inc_list = [
            tg.env.freespace_3d.PlaneWave(
                e0p=0.0, e0s=1.0, inc_angle=torch.pi / 3
            ),  # s-polarization, incidence from some angle
        ]

        # - structure
        step = 10.0
        mat_struct = tg.materials.MatConstant(eps=10.0 + 0.5j)
        struct1 = tg.struct.StructDiscretizedCubic3D(
            tg.struct.volume.cuboid(l=5, w=10, h=3),
            step,
            mat_struct,
            radiative_correction=False,
        )

        self.sim = tg.simulation.Simulation(
            structures=[struct1],
            environment=env,
            illumination_fields=e_inc_list,
            wavelengths=self.wavelengths,
        )

    def test_farfield_vs_nearfield(self):
        for device in self.devices:
            self.sim.set_device(device)
            self.sim.run(verbose=False, progress_bar=False)

            sf_spherical = tg.tools.geometry.coordinate_map_2d_spherical(
                n_phi=18, n_teta=9
            )

            # far-field integration
            ff_res = tg.postproc.fields.ff(
                self.sim, self.wavelengths[0], sf_spherical, progress_bar=False
            )
            r_ff = ff_res["sca"]
            I_sc_ff_int = r_ff.get_integrated_efield_intensity(progress_bar=False)

            # near-field integration in far-zone (slower)
            nf_res = tg.postproc.fields.nf(
                self.sim, self.wavelengths[0], sf_spherical, progress_bar=False
            )
            r_nf = nf_res["sca"]
            I_sc_nf_int = r_nf.get_integrated_efield_intensity(progress_bar=False)

            # print(I_sc_ff_int, I_sc_nf_int)

            torch.testing.assert_close(I_sc_nf_int, I_sc_ff_int, rtol=1e-5, atol=0.1)

            if self.verbose:
                print("  - {}: far-field vs near-field test passed.".format(device))

    def test_farfield_vs_crosssections(self):
        for device in self.devices:
            self.sim.set_device(device)
            self.sim.run(verbose=False, progress_bar=False)

            # cross-section
            res_cs = tg.postproc.crosssect.total(
                self.sim, wavelength=self.wavelengths[0], progress_bar=False
            )
            scs = res_cs["scs"]

            # far-field integration
            sf_spherical = tg.tools.geometry.coordinate_map_2d_spherical(
                n_phi=360, n_teta=180
            )
            ff_res = tg.postproc.fields.ff(self.sim, self.wavelengths[0], sf_spherical, progress_bar=False)
            r_ff = ff_res["sca"]
            I_sc_int = r_ff.get_integrated_efield_intensity(progress_bar=False)

            # print(scs, I_sc_int)
            torch.testing.assert_close(scs, I_sc_int, rtol=2e-3, atol=15)

            if self.verbose:
                print(
                    "  - {}: far-field integration vs SCS test passed.".format(device)
                )


# %%

# %%
if __name__ == "__main__":
    print("testing fields postprocessing...")
    torch.set_printoptions(precision=7)
    unittest.main(argv=["first-arg-is-ignored"], exit=False)

    # t = TestFieldCalculations()
    # t.test_farfield_vs_crosssections()



# encoding=utf-8
# %%
import unittest

import torch

import torchgdm as tg


class TestComplexGreen3D(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print(
                "testing Green's function postprocessing with discretized 3D-simulation..."
            )

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # --- setup a test case
        # - environment
        self.wavelength = 550.0
        self.r_source = torch.as_tensor([[-50, 0, 40], [10, 0, -40]])
        self.r_probe = tg.tools.geometry.coordinate_map_2d_square(100, 5, r3=50)

        # --- setup test case simulation
        # - environment
        eps_env = 1.33
        mat_env = tg.materials.MatConstant(eps_env)
        env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=mat_env)

        # - dummy illumination field
        wavelengths = torch.tensor([self.wavelength])
        e_inc_dummy = tg.env.freespace_3d.NullField()

        # - first structure: volume discretization
        l = w = h = 3
        step = 20.0
        mat_struct = tg.materials.MatConstant(eps=14.0)

        struct_mesh = tg.struct.StructDiscretizedCubic3D(
            tg.struct.volume.cuboid(l, w, h), step, mat_struct
        )
        struct_mesh.set_center_of_mass([0, 0, 0])

        # - full simulation
        self.sim = tg.simulation.Simulation(
            structures=[struct_mesh],
            environment=env,
            illumination_fields=[e_inc_dummy],
            wavelengths=wavelengths,
        )

    def test_green_implemenetations_equality(self):
        for device in self.devices:
            self.sim.set_device(device)

            G_6x6_1 = tg.postproc.green._G_direct(
                self.sim,
                r_probe=self.r_probe,
                r_source=self.r_source,
                wavelength=self.wavelength,
                progress_bar=False,
            )
            G_results = tg.postproc.green.G(
                self.sim,
                wavelength=self.wavelength,
                r_probe=self.r_probe,
                r_source=self.r_source,
                progress_bar=False,
                verbose=False,
            )
            G_6x6_2 = G_results["G_6x6"]

            torch.testing.assert_close(G_6x6_1, G_6x6_2)

    def test_green_far_from_structures(self):
        for device in self.devices:
            self.sim.set_device(device)

            # source and probe far from structure
            dx, dy = 10000, 15000
            r_source = torch.as_tensor(
                [dx, dy, 35], dtype=torch.float32, device=device
            )
            r_probe = tg.tools.geometry.coordinate_map_2d_square(
                500, 5, delta1=dx, delta2=dy, r3=40
            )["r_probe"].to(device)

            G_results = tg.postproc.green.G(
                self.sim,
                wavelength=self.wavelength,
                r_probe=r_probe,
                r_source=r_source,
                progress_bar=False,
                verbose=False,
            )
            G_sim = G_results["G_6x6"][0]  # first source position (of one...)

            G_vac = self.sim.environment.get_G_6x6(r_probe, r_source, self.wavelength)

            torch.testing.assert_close(G_sim, G_vac, rtol=1e-7, atol=1e-7)

    def test_green_output_consistency(self):
        for device in self.devices:
            self.sim.set_device(device)

            G_results = tg.postproc.green.G(
                self.sim,
                wavelength=self.wavelength,
                r_probe=self.r_probe,
                r_source=self.r_source,
                progress_bar=False,
                verbose=False,
            )

            torch.testing.assert_close(
                G_results["G_6x6"][..., :3, :3], G_results["G_Ep"]
            )
            torch.testing.assert_close(
                G_results["G_6x6"][..., 3:, 3:], G_results["G_Hm"]
            )
            torch.testing.assert_close(
                G_results["G_6x6"][..., 3:, :3], G_results["G_Hp"]
            )
            torch.testing.assert_close(
                G_results["G_6x6"][..., :3, 3:], G_results["G_Em"]
            )

    def test_green_on_sample(self):
        for device in self.devices:
            self.sim.set_device(device)

            G_results = tg.postproc.green.G(
                self.sim,
                wavelength=self.wavelength,
                r_probe=self.r_probe,
                r_source=self.r_source,
                progress_bar=False,
                verbose=False,
            )
            G_6x6 = G_results["G_6x6"][1, [7, 15]]
            # print(G_6x6)

            G_6x6_ref = torch.as_tensor(
                [
                    [
                        [
                            -1.0886665e-06 + 3.4371817e-07j,
                            1.9067019e-07 + 1.4734553e-08j,
                            -3.1135608e-07 - 2.2034816e-07j,
                            7.8117388e-09 - 7.9646398e-08j,
                            -9.4577649e-07 + 9.5512132e-07j,
                            -4.0747094e-07 + 9.6144527e-07j,
                        ],
                        [
                            4.4389697e-07 + 3.4053393e-08j,
                            -1.7907624e-08 + 4.4039223e-07j,
                            -3.1159952e-06 - 5.4947202e-07j,
                            1.1503031e-06 - 1.7255561e-06j,
                            3.4392386e-08 - 1.6531759e-07j,
                            1.6279989e-07 - 2.4089610e-07j,
                        ],
                        [
                            -4.3981481e-07 - 1.1056457e-07j,
                            -9.5553878e-07 - 2.3271329e-07j,
                            2.5345098e-06 + 1.8447427e-06j,
                            1.7052392e-07 + 1.4413263e-07j,
                            -1.4393464e-07 + 2.4122363e-07j,
                            -4.3346702e-08 + 2.1297481e-07j,
                        ],
                        [
                            -4.5305676e-08 + 2.0207312e-07j,
                            6.2416444e-07 - 4.5764466e-07j,
                            8.8299618e-07 - 1.8992670e-06j,
                            -1.5808587e-06 + 4.8191225e-07j,
                            2.4335304e-07 + 4.8928911e-08j,
                            -3.7928774e-07 - 7.7906037e-08j,
                        ],
                        [
                            -7.2510016e-07 + 1.1805100e-06j,
                            -1.2150139e-09 + 1.2138926e-08j,
                            -1.8058941e-09 - 5.1080988e-08j,
                            1.9038019e-07 + 1.0075015e-08j,
                            -8.2802069e-07 + 5.4201143e-07j,
                            -1.5086823e-06 - 1.0375794e-07j,
                        ],
                        [
                            -3.1773121e-07 + 4.1915200e-08j,
                            7.0668015e-08 - 6.8113444e-08j,
                            1.3259778e-07 - 4.7104811e-07j,
                            -3.4099307e-07 - 5.0336610e-08j,
                            -1.9207364e-06 - 3.6182652e-07j,
                            1.6083627e-06 + 1.0215560e-06j,
                        ],
                    ],
                    [
                        [
                            -2.6855030e-08 + 3.6664164e-07j,
                            -1.8608105e-07 - 7.7725915e-08j,
                            -1.0681069e-06 - 4.9962375e-07j,
                            -7.1662441e-08 + 2.0746609e-07j,
                            -8.4293180e-07 + 5.1455959e-07j,
                            3.0324927e-07 - 2.3444996e-07j,
                        ],
                        [
                            -2.2717860e-07 - 8.7862361e-08j,
                            -4.0295907e-07 + 1.5908353e-07j,
                            5.7338218e-07 + 1.7160788e-07j,
                            7.4529407e-07 - 2.1550466e-07j,
                            5.8602843e-08 - 1.8308809e-07j,
                            7.2966486e-07 - 5.4695380e-07j,
                        ],
                        [
                            -4.3967685e-07 - 2.8341822e-07j,
                            1.9003964e-07 + 9.7208570e-08j,
                            -3.8460701e-07 + 9.4305631e-07j,
                            -2.6108958e-07 + 1.2258243e-07j,
                            -6.1210079e-07 + 2.5809229e-07j,
                            8.2391143e-09 - 1.9775118e-08j,
                        ],
                        [
                            4.7302962e-08 - 1.3970885e-07j,
                            4.5206298e-07 - 2.6617460e-07j,
                            -5.6532929e-07 + 4.4056014e-07j,
                            -1.3186141e-07 + 4.3942893e-07j,
                            -3.5422531e-07 - 1.1052117e-07j,
                            -6.6941493e-07 - 2.0041207e-07j,
                        ],
                        [
                            -4.2451379e-07 + 8.1479044e-08j,
                            -1.5656733e-08 + 8.7590564e-08j,
                            -1.1492109e-06 + 8.7112801e-07j,
                            -3.4184518e-07 - 8.9863313e-08j,
                            -7.7181198e-07 + 2.4102647e-07j,
                            2.9585524e-07 + 7.4918162e-08j,
                        ],
                        [
                            1.9843461e-07 - 7.7167357e-08j,
                            3.8585475e-07 - 1.2949488e-07j,
                            -8.5867455e-08 + 1.0114601e-07j,
                            -8.7262913e-07 - 4.8787035e-07j,
                            4.0747955e-07 + 2.2172523e-07j,
                            -2.8789901e-07 + 5.4970485e-07j,
                        ],
                    ],
                ],
                dtype=torch.complex64,
                device=device,
            )
            torch.testing.assert_close(G_6x6, G_6x6_ref)


class TestLDOS3D(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing LDOS with discretized 3D-simulation...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # --- setup a test case
        # - environment
        self.wavelength = 550.0
        self.r_probe = tg.tools.geometry.coordinate_map_2d_square(100, 5, r3=50)

        # --- setup test case simulation
        # - environment
        eps_env = 1.33
        mat_env = tg.materials.MatConstant(eps_env)
        env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=mat_env)

        # - dummy illumination field
        wavelengths = torch.tensor([self.wavelength])
        e_inc_dummy = tg.env.freespace_3d.NullField()

        # - first structure: volume discretization
        l = w = h = 3
        step = 20.0
        mat_struct = tg.materials.MatConstant(eps=14.0)

        struct_mesh = tg.struct.StructDiscretizedCubic3D(
            tg.struct.volume.cuboid(l, w, h), step, mat_struct
        )
        struct_mesh.set_center_of_mass([0, 0, 0])

        # - full simulation
        self.sim = tg.simulation.Simulation(
            structures=[struct_mesh],
            environment=env,
            illumination_fields=[e_inc_dummy],
            wavelengths=wavelengths,
        )

    def test_ldos_vs_green(self):
        for device in self.devices:
            self.sim.set_device(device)

            ldos_results = tg.postproc.green.ldos(
                self.sim,
                r_probe=self.r_probe,
                wavelength=self.wavelength,
                progress_bar=False,
            )

            G_ii_1 = ldos_results["G_ii"]

            G_results = tg.postproc.green.G(
                self.sim,
                wavelength=self.wavelength,
                r_probe=self.r_probe,
                r_source=self.r_probe,
                progress_bar=False,
                verbose=False,
            )

            G_6x6 = G_results["G_6x6"]

            G_ii_2 = G_6x6[
                torch.arange(G_6x6.shape[0]), torch.arange(G_6x6.shape[1]), ...
            ]

            torch.testing.assert_close(G_ii_1, G_ii_2)

    def test_ldos_far_from_structures(self):
        for device in self.devices:
            self.sim.set_device(device)

            # relative ldos far from structure (should be one)
            dx, dy = 10000, 15000
            r_probe = tg.tools.geometry.coordinate_map_2d_square(
                1000, 5, delta1=dx, delta2=dy, r3=50
            )["r_probe"].to(device)

            ldos_results = tg.postproc.green.ldos(
                self.sim,
                r_probe=r_probe,
                wavelength=self.wavelength,
                progress_bar=False,
            )

            ldos_p = ldos_results["ldos_partial"]
            ldos_e = ldos_results["ldos_e"]
            ldos_m = ldos_results["ldos_m"]
            
            ones_p = torch.ones_like(ldos_p)
            ones = torch.ones_like(ldos_e)

            torch.testing.assert_close(ldos_p, ones_p)
            torch.testing.assert_close(ldos_e, ones)
            torch.testing.assert_close(ldos_m, ones)

    def test_ldos_output_consistency(self):
        for device in self.devices:
            self.sim.set_device(device)

            ldos_results = tg.postproc.green.ldos(
                self.sim,
                r_probe=self.r_probe,
                wavelength=self.wavelength,
                progress_bar=False,
            )

            ldos_p = ldos_results["ldos_partial"]
            ldos_e = ldos_results["ldos_e"]
            ldos_m = ldos_results["ldos_m"]

            torch.testing.assert_close(ldos_e, 1 + torch.sum(ldos_p[:, :3] - 1, dim=-1))
            torch.testing.assert_close(ldos_m, 1 + torch.sum(ldos_p[:, 3:] - 1, dim=-1))

    def test_ldos_on_sample(self):
        for device in self.devices:
            self.sim.set_device(device)

            ldos_results = tg.postproc.green.ldos(
                self.sim,
                r_probe=self.r_probe,
                wavelength=self.wavelength,
                progress_bar=False,
            )
            ldos_p = ldos_results["ldos_partial"][::4]
            # print(ldos_p)

            ldos_p_ref = torch.as_tensor(
                [
                    [1.0114026, 1.0114026, 0.9800825, 1.0430276, 1.0430276, 1.0649197],
                    [1.0114025, 1.0114026, 0.9800825, 1.0430276, 1.0430276, 1.0649198],
                    [1.1013016, 1.1013016, 1.1013016, 1.1343262, 1.1343262, 1.1343262],
                    [0.5307579, 0.5307579, 2.5567667, 1.3563178, 1.3563179, 1.1544354],
                    [1.1013016, 1.1013014, 1.1013016, 1.1343262, 1.1343262, 1.1343262],
                    [1.0114026, 1.0114025, 0.9800825, 1.0430276, 1.0430276, 1.0649197],
                    [1.0114026, 1.0114026, 0.9800825, 1.0430276, 1.0430276, 1.0649198],
                ],
                dtype=torch.float32,
                device=device,
            )
            torch.testing.assert_close(ldos_p, ldos_p_ref)


# %%

# %%

# %%
if __name__ == "__main__":
    print(
        "testing Green's function calc. in complex environment structure Greens tensor and LDOS calculation..."
    ), torch.set_printoptions(precision=7)
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# -*- coding: utf-8 -*-
"""
illumination fields for torchgdm
"""
import warnings

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device
from torchgdm.env.base_classes import IlluminationfieldBase

# make 2D compatible fields available
from torchgdm.env.freespace_3d.inc_fields import NullField as _nullfield3D
from torchgdm.env.freespace_3d.inc_fields import PlaneWave as _planewave3D


class NullField(_nullfield3D):
    """Zero-Field

    all additional kwargs are ignored

    """

    __name__ = "null field - 2D"
    n_dim = 2


class PlaneWave(_planewave3D):
    """2D plane wave in homogeneous environment"""

    __name__ = "free space plane wave 2D"

    def __init__(
        self,
        e0s=0.0,
        e0p=1.0,
        inc_angle=0.0,
        inc_plane="xz",
        phase_e0s=0.0,
        phase_global=0,
        device: torch.device = None,
    ):
        """plane wave in homogeneous 2D environment

        The polarization state can be defined in different ways. The s- and p-polarization
        complex amplitudes can be defined with a relative phase using. Alternatively a
        relative phase (in rad) can be given for the s-polarization component.

        Args:
            e0s (complex, optional): s-polarization amplitude factor. Defaults to 0.0.
            e0p (complex, optional): s-polarization amplitude factor. Defaults to 1.0.
            inc_angle (float, optional): incident angle in rad. 0:k=-ez, pi:k=+ez. Defaults to 0.0.
            inc_plane (str, optional): Either "xz" or "yz". Defaults to "xz".
            phase_e0s (float, optional): additional phase of s-amplitude vector (in rad). Defaults to 0.0.
            phase_global (int, optional): additional global phase (in rad). Defaults to 0.
            device (torch.device, optional): Defaults to "cpu".
        """
        super().__init__(
            e0s=e0s,
            e0p=e0p,
            inc_angle=inc_angle,
            phase_e0s=phase_e0s,
            phase_global=phase_global,
            inc_plane=inc_plane,
            device=device,
        )
        self.n_dim = 2


class ElectricLineDipole(IlluminationfieldBase):
    """2D electric line-dipole in homogeneous environment"""

    __name__ = "electric line dipole"

    def __init__(
        self,
        r_source: torch.Tensor,
        p_source: torch.Tensor,
        device: torch.device = None,
    ):
        """2D electric line-dipole in homogeneous environment

        Args:
            r_source (complex torch.Tensor): 2D position of line dipole as [x, z] or [x, 0, z] (y: infinite axis)
            p_source (complex torch.Tensor): amplitude vector (3D) of the source's electric dipole moment as [p_x, p_y, p_z]
            device (torch.device, optional): Defaults to "cpu".
        """
        super().__init__(device=device)
        self.n_dim = 2

        self.r_source = torch.as_tensor(r_source, dtype=DTYPE_FLOAT, device=device)
        self.p_source = torch.as_tensor(p_source, dtype=DTYPE_COMPLEX, device=device)

        assert len(self.r_source.shape) == 1
        if len(self.r_source) == 2:
            self.r_source = torch.tensor([self.r_source[0], 0, self.r_source[1]])
        assert len(r_source) == 3 and len(p_source) == 3
        assert self.r_source[1] == 0

    def set_device(self, device):
        super().set_device(device)
        self.r_source = self.r_source.to(device=device)
        self.p_source = self.p_source.to(device=device)

    def __repr__(self, verbose=False):
        out_str = " ------ {} -------".format(self.__name__)
        out_str += "\n - location (nm): {}".format(
            torch.round(self.r_source, decimals=1)
        )
        out_str += "\n - amplitude: {} ".format(self.p_source)

        return out_str

    def get_info(self):
        return dict(
            name=self.__name__,
            r_source=self.r_source,
            p_source=self.p_source,
        )

    def get_x(self):
        """return x position of line dipole"""
        return self.r_source.squeeze()[0]

    def get_y(self):
        """return y position of line dipole"""
        # this is 0 by definition
        return self.r_source.squeeze()[1]

    def get_z(self):
        """return z position of line dipole"""
        return self.r_source.squeeze()[2]

    def get_efield(self, r_probe: torch.Tensor, wavelength: float, environment):
        """evaluate the illumination's electric field

        Args:
            r_probe (torch.Tensor): evaluation position(s)
            wavelength (float): in nm
            environment (:class:`torchgdm.env.base_classes.EnvironmentBase`): Evaluation environment

        Returns:
            torch.Tensor: E-field(s) at evaluation position(s)
        """
        r_probe = torch.as_tensor(r_probe, dtype=DTYPE_FLOAT, device=self.device)
        if len(r_probe.shape) == 1:
            r_probe = r_probe.unsqueeze(0)

        G_Ep = environment.get_G_Ep(r_probe, self.r_source.unsqueeze(0), wavelength)
        Ep = torch.matmul(G_Ep, self.p_source.unsqueeze(-1))[..., 0]
        return Ep

    def get_hfield(self, r_probe: torch.Tensor, wavelength: float, environment):
        """evaluate the illumination's magnetic field

        Args:
            r_probe (torch.Tensor): evaluation position(s)
            wavelength (float): in nm
            environment (:class:`torchgdm.env.base_classes.EnvironmentBase`): Evaluation environment

        Returns:
            torch.Tensor: H-field(s) at evaluation position(s)
        """
        r_probe = torch.as_tensor(r_probe, dtype=DTYPE_FLOAT, device=self.device)
        if len(r_probe.shape) == 1:
            r_probe = r_probe.unsqueeze(0)

        G_Hp = environment.get_G_Hp(r_probe, self.r_source.unsqueeze(0), wavelength)
        Hp = torch.matmul(G_Hp, self.p_source.unsqueeze(-1))[..., 0]
        return Hp


class MagneticLineDipole(IlluminationfieldBase):
    """2D magnetic line-dipole in homogeneous environment"""

    __name__ = "magnetic line dipole"

    def __init__(
        self,
        r_source: torch.Tensor,
        m_source: torch.Tensor,
        device: torch.device = None,
    ):
        """2D magnetic line-dipole in homogeneous environment

        Args:
            r_source (complex torch.Tensor): 2D position of line dipole as [x, z] or [x, 0, z] (y: infinite axis)
            m_source (complex torch.Tensor): amplitude vector (3D) of the source's magnetic dipole moment as [p_x, p_y, p_z]
            device (torch.device, optional): Defaults to "cpu".
        """
        super().__init__(device=device)
        self.n_dim = 2

        self.r_source = torch.as_tensor(r_source, dtype=DTYPE_FLOAT, device=device)
        self.m_source = torch.as_tensor(m_source, dtype=DTYPE_COMPLEX, device=device)

        assert len(self.r_source.shape) == 1
        if len(self.r_source) == 2:
            self.r_source = torch.tensor([self.r_source[0], 0, self.r_source[1]])
        assert len(r_source) == 3 and len(m_source) == 3
        assert self.r_source[1] == 0

    def set_device(self, device):
        super().set_device(device)
        self.r_source = self.r_source.to(device=device)
        self.m_source = self.m_source.to(device=device)

    def __repr__(self, verbose=False):
        out_str = " ------ {} -------".format(self.__name__)
        out_str += "\n - location (nm): {}".format(
            torch.round(self.r_source, decimals=1)
        )
        out_str += "\n - amplitude: {} ".format(self.m_source)

        return out_str

    def get_info(self):
        return dict(
            name=self.__name__,
            r_source=self.r_source,
            p_source=self.m_source,
        )

    def get_x(self):
        """return x position of line dipole"""
        return self.r_source.squeeze()[0]

    def get_y(self):
        """return y position of line dipole"""
        # this is 0 by definition
        return self.r_source.squeeze()[1]

    def get_z(self):
        """return z position of line dipole"""
        return self.r_source.squeeze()[2]

    def get_efield(self, r_probe: torch.Tensor, wavelength: float, environment):
        """evaluate the illumination's electric field

        Args:
            r_probe (torch.Tensor): evaluation position(s)
            wavelength (float): in nm
            environment (:class:`torchgdm.env.base_classes.EnvironmentBase`): Evaluation environment

        Returns:
            torch.Tensor: E-field(s) at evaluation position(s)
        """
        r_probe = torch.as_tensor(r_probe, dtype=DTYPE_FLOAT, device=self.device)
        if len(r_probe.shape) == 1:
            r_probe = r_probe.unsqueeze(0)

        G_Em = environment.get_G_Em(r_probe, self.r_source.unsqueeze(0), wavelength)
        Ep = torch.matmul(G_Em, self.m_source.unsqueeze(-1))[..., 0]
        return Ep

    def get_hfield(self, r_probe: torch.Tensor, wavelength: float, environment):
        """evaluate the illumination's magnetic field

        Args:
            r_probe (torch.Tensor): evaluation position(s)
            wavelength (float): in nm
            environment (:class:`torchgdm.env.base_classes.EnvironmentBase`): Evaluation environment

        Returns:
            torch.Tensor: H-field(s) at evaluation position(s)
        """
        r_probe = torch.as_tensor(r_probe, dtype=DTYPE_FLOAT, device=self.device)
        if len(r_probe.shape) == 1:
            r_probe = r_probe.unsqueeze(0)

        G_Hm = environment.get_G_Hm(r_probe, self.r_source.unsqueeze(0), wavelength)
        Hp = torch.matmul(G_Hm, self.m_source.unsqueeze(-1))[..., 0]
        return Hp



# encoding=utf-8
"""environments classes and illumination fields


An environment class implements:
  
    - Green's tensors, describing environments and their boundary conditions.  
    - illumination fields specific for the environment


.. currentmodule:: torchgdm.env


Environment classes
-------------------

.. autosummary::
   :toctree: generated/

   EnvHomogeneous2D
   EnvHomogeneous3D
   EnvironmentBase


Illumination classes
--------------------

Only the illumination base class is defined here.

.. autosummary::
   :toctree: generated/

   IlluminationfieldBase


The illumination classes are defined in the subpackage of each environment.

.. autosummary::
   :toctree: generated/

   freespace_2d
   freespace_3d
  
"""
from .base_classes import IlluminationfieldBase, EnvironmentBase

from . import freespace_2d
from . import freespace_3d

from .freespace_2d import EnvHomogeneous2D
from .freespace_3d import EnvHomogeneous3D



# -*- coding: utf-8 -*-
"""3D visualization tools based on pyvista

Requires `pyvista`. For interactive use in jupyter, requires additionally: `trame`, `trame-vtk`, `trame-vuetify`, `ipywidgets`. Install with:

.. code-block:: bash

     pip install 'jupyterlab>=3' ipywidgets 'pyvista[all,trame]'
 

*Note:* 3D visualizations of simulations or fields can be readily accessed through the :class:`torchgdm.Simulation` and :class:`torchgdm.Field` classes.


.. autosummary::
   :toctree: generated/
    
    structure
    vectorfield
    vectorfield_inside

"""
from . import geo3d
from . import vec3d

from .geo3d import structure
from .vec3d import vectorfield, vectorfield_inside



# -*- coding: utf-8 -*-
"""3D effective point polarizabilities

.. currentmodule:: torchgdm.struct.point

.. autosummary::
   :toctree: generated/
   
   StructEffPola3D
   StructMieSphereEffPola3D
"""
from .pola import StructEffPola3D
from .pola import StructMieSphereEffPola3D
from . import pola



# -*- coding: utf-8 -*-
"""2D surface discretizations

.. currentmodule:: torchgdm.struct.surface_2d

Classes
-------

.. autosummary::
   :toctree: generated/
   
   StructDiscretized2D
   StructDiscretizedSquare2D


Functions
---------

.. autosummary::
   :toctree: generated/
   
   extract_eff_pola_via_propagation


Geometries
----------

.. autosummary::
   :toctree: generated/
   
   square
   rectangle
   circle
   split_ring
   triangle_equilateral


Discretizer functions
---------------------

.. autosummary::
   :toctree: generated/
   
   discretizer_square

"""
from .pola import StructDiscretized2D, StructDiscretizedSquare2D
from .pola import extract_eff_pola_via_propagation
from . import pola
from . import geometries

from .geometries import discretizer_square

from .geometries import square
from .geometries import rectangle
from .geometries import circle
from .geometries import split_ring
from .geometries import triangle_equilateral



# -*- coding: utf-8 -*-
"""
handling of linear equation systems (coupling)    
"""
# %%
from collections.abc import Callable
import time
import warnings

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import purge_mem, get_default_device
from torchgdm.tools.batch import batched


def _reduce_dimensions(Q: torch.Tensor):
    """expand dimensions of Q from (A1, A2, N, N) to (A1*N, A2*N)

    Args:
        Q (torch.Tensor): 4D tensor of shape (A1, A2, N, N), containing an array of (N, N) tensors

    Returns:
        torch.Tensor: tensor of shape (A1*N, A2*N) with conserved global structure of Q
    """
    Q_red = Q.permute(0, 2, 1, 3)
    dim_Q1 = Q_red.shape[0] * Q_red.shape[1]
    dim_Q2 = Q_red.shape[2] * Q_red.shape[3]

    # note: this is not memory efficient as it copies the full matrix...
    Q_red = Q_red.contiguous().view(dim_Q1, dim_Q2)
    return Q_red


def _expand_dimensions(Q: torch.Tensor, N: int = 6):
    """expand dimensions of Q from (A1*N, A2*N) to (A1, A2, N, N)

    Args:
        Q (torch.Tensor): 4D tensor of shape (A1, A2, N, N), containing an array of (N, N) tensors
        N (int): size of rank 2 sub-tensors. Defaults to 6

    Returns:
        torch.Tensor: tensor of shape (A1, A2, N, N) with conserved global structure of Q
    """
    dim_Q1 = Q.shape[0] // N
    dim_Q2 = Q.shape[1] // N

    # note: this is not memory efficient as it copies the full matrix...
    Q_exp = Q.contiguous().view(dim_Q1, N, dim_Q2, N)
    Q_exp = Q_exp.permute(0, 2, 1, 3)

    return Q_exp


@batched("f0")
def _batched_lu_solve(LU, pivots, f0=None):
    if f0 is None:
        raise ValueError("'f0' must be given.")
    return torch.linalg.lu_solve(LU, pivots, f0)


# --- linear system base class
class LinearSystemBase:
    """torchgdm base class for linearsystem, describing and solving the coupling"""

    __name__ = "linearsystem base class"

    def __init__(self, device: torch.device = None):
        if device is None:
            self.device = get_default_device()
        else:
            self.device = device

    def set_device(self, device):
        """move all tensors of the class to device"""
        self.device = device

    def solve(self):
        """solve coupling for illumination(s)"""
        raise NotImplementedError("`solve` not implemented with this linear system.")

    def get_generalized_propagator(self):
        """get the generalized propagator"""
        raise NotImplementedError(
            "Generalized propoagator not available with this linear system."
        )


class LinearSystemFullInverse(LinearSystemBase):
    """Solve coupling by full inversion

    This linear system solver class is a non-optimized reference implementation for testing.
    It uses (6x6) polarizabilities for all dipoles, which is non memory efficient in most cases.
    """

    def __init__(self, device: torch.device = None):
        """Solve coupling by full inversion

        Memory-inefficient test implementation. Uses 6x6 polarizabilities for every positio.

        Args:
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).
        """
        super().__init__(device=device)

    def set_device(self, device):
        """move all tensors of the class to device"""
        super().set_device(device)

    # --- functions to setup interaction systems
    def _get_full_interaction_matrix_G_tensors(
        self, positions: torch.Tensor, G_func: Callable, wavelength: float
    ) -> torch.Tensor:
        # evaluate Green's tensors for each couple of points
        # iteract_G_NxN is of shape (N,N,3,3)
        interact_G_NxN = G_func(
            positions.unsqueeze(1),
            positions.unsqueeze(0),
            wavelength=wavelength,
        )
        return interact_G_NxN

    def _get_full_Gdotalpha(
        self,
        positions: torch.Tensor,
        polarizabilities: torch.Tensor,
        self_terms: torch.Tensor,
        G_func: Callable,
        wavelength: float,
    ) -> torch.Tensor:
        """calc. invertible interaction matrix $(G.alpha)$, selfterms on diagonal"""
        N = len(positions)

        interact_NxN = self._get_full_interaction_matrix_G_tensors(
            positions, G_func, wavelength
        )

        # put self-term matrices on diagonal (at R1[i]==R2[i])
        eye_indices = torch.arange(N, device=self.device)
        interact_NxN[eye_indices, eye_indices, :, :] = self_terms.unsqueeze(0)

        # multiply by polarizabilities
        interact_NxN = torch.matmul(interact_NxN, polarizabilities.unsqueeze(0))

        # reshape (N,N,6,6) to (6N, 6N) matrix (or (3N,3N) if electric-only)
        interact_6Nx6N = _reduce_dimensions(interact_NxN)

        return interact_6Nx6N

    def get_interact(self, sim, wavelength: float) -> torch.Tensor:
        """get the full interaction block (E, H) of the linear system for a list of structures

        containes all direct and cross-contributions (E-E, H-H, E-H, H-E)
        """
        # calc. interactions (G.alpha)
        all_positions = sim.get_all_positions()

        interact = self._get_full_Gdotalpha(
            positions=all_positions,
            polarizabilities=sim._get_all_polarizabilitites_6x6(wavelength),
            self_terms=sim._get_all_selfterms_6x6(wavelength),
            G_func=sim.environment.get_G_6x6,
            wavelength=wavelength,
        )

        # invertible matrix block is (1 - G.alpha), which adds the illumination field
        interact = torch.eye(len(interact), device=self.device) - interact

        return interact

    def solve(
        self,
        sim,
        wavelength: float,
        batch_size=32,
        verbose: int = 1,
    ) -> torch.Tensor:
        """solve system of coupled structures for all illuminations

        Args:
            sim (SimulationBase): simulation container
            wavelength (float): wavelength (nm)
            batch_size (int, optional): Nr of incident fields evaluated in parallel. Defaults to 32.
            verbose (int, optional): wether to print status info. Defaults to 1.

        Returns:
            torch.Tensor: list of internal fields, one for each illumination. Shape: (N_e0, nr of pos, 3)
        """
        if verbose:
            t_start = time.time()
            print("setup... ", end="")

        # - setup interaction system
        interact = self.get_interact(sim, wavelength)
        if verbose:
            t_interact = time.time()
            print("done in {:.2f}s.\nLU... ".format(t_interact - t_start), end="")
            print("done in {:.2f}s.\nLU... ".format(t_interact - t_start), end="")

        # - inversion
        # LU, pivots, info = torch.linalg.lu_factor_ex(interact)  # experimental version, faster on GPU
        LU, pivots = torch.linalg.lu_factor(interact)
        if verbose:
            t_lu = time.time()
            print("   done in {:.2f}s.\nE0... ".format(t_lu - t_interact), end="")

        # - evaluate zero order fields
        f0 = sim._get_all_e0_h0(wavelength)
        f0 = f0.view(len(f0), -1, 1)
        if verbose:
            t_e0 = time.time()
            print("   done in {:.2f}s.\nsolve... ".format(t_e0 - t_lu), end="")

        # - solve for RHS vectors
        f_inside = _batched_lu_solve(LU, pivots, f0=f0, batch_size=batch_size)
        eh_inside = f_inside.view(len(f_inside), -1, 6)  # conv 6N to N 6-tuples (E,H)
        e_inside, h_inside = torch.chunk(eh_inside, 2, dim=2)  # separate E and H fields

        if verbose:
            t_solve = time.time()
            print("done in {:.2f}s.".format(t_solve - t_e0))

        return e_inside, h_inside

    def get_generalized_propagator(self, sim, wavelength: float) -> torch.Tensor:
        interact = self.get_interact(sim, wavelength)
        K = torch.linalg.inv(interact)
        return K


class LinearSystemFullMoreMemoryEfficient(LinearSystemFullInverse):
    """Solve coupling by full inversion

    This linear system solver couples only actually polarizable terms.
    """

    def __init__(self, device: torch.device = None):
        """Solve coupling by full inversion

        This linear system solver couples only actually polarizable terms.

        Args:
            device (torch.device, optional): Defaults to "cpu" (can be changed by :func:`torchgdm.use_cuda`).
        """
        super().__init__(device=device)

    def set_device(self, device):
        """move all tensors of the class to device"""
        super().set_device(device)

    # --- select proper coupling for structure pair
    def get_struct_pair_coupling_config(
        self,
        struct1,
        struct2,
        sim,
        verbose=False,
    ):
        """get evaluation functions for required coupling between two structures

        Concerns Greens tensor, polarizability and self-term evaluation

        Args:
            struct1 (structBase): target of scattered light
            struct2 (structBase): origin of light scattering
            sim (SimulationBase): simulation descriptor

        Returns:
            callable: Green's tensor calculation function
            callable: polarizability calculation function
            callable: self-term evaluator, only returned for equal coupling of st1 and st2. default: None
        """

        _s1T = struct1.evaluation_terms  # observer structure (target)
        _s2T = struct2.evaluation_terms  # emitter structure (origin)
        selfterm_func = (
            None  # default: two different structures --> no self-terms required
        )

        # --- pure observer, pure source
        # e <- e
        if ("E" in _s1T and "H" not in _s1T) and ("E" in _s2T and "H" not in _s2T):
            G_func = sim.environment.get_G_Ep
            alpha_func = struct2.get_polarizability_pE
            selfterm_func = struct2.get_selfterm_pE
        # h <- h
        elif ("E" not in _s1T and "H" in _s1T) and ("E" not in _s2T and "H" in _s2T):
            G_func = sim.environment.get_G_Hm
            alpha_func = struct2.get_polarizability_mH
            selfterm_func = struct2.get_selfterm_mH
        # h <- e
        elif ("E" not in _s1T and "H" in _s1T) and ("E" in _s2T and "H" not in _s2T):
            G_func = sim.environment.get_G_Hp
            alpha_func = struct2.get_polarizability_pE
        # e <- h
        elif ("E" in _s1T and "H" not in _s1T) and ("E" not in _s2T and "H" in _s2T):
            G_func = sim.environment.get_G_Em
            alpha_func = struct2.get_polarizability_mH

        # --- mixed observer, pure source
        # e,m <- e
        elif ("E" in _s1T and "H" in _s1T) and ("E" in _s2T and "H" not in _s2T):
            G_func = sim.environment.get_G_6x6
            alpha_func = struct2.get_polarizability_pmE_6x3
        # e,m <- m
        elif ("E" in _s1T and "H" in _s1T) and ("E" not in _s2T and "H" in _s2T):
            G_func = sim.environment.get_G_6x6
            alpha_func = struct2.get_polarizability_pmH_6x3

        # --- pure observer, mixed source
        # e <- e,m
        elif ("E" in _s1T and "H" not in _s1T) and ("E" in _s2T and "H" in _s2T):
            G_func = sim.environment.get_G_Epm_3x6
            alpha_func = struct2.get_polarizability_6x6
        # m <- e,m
        elif ("E" not in _s1T and "H" in _s1T) and ("E" in _s2T and "H" in _s2T):
            G_func = sim.environment.get_G_Hpm_3x6
            alpha_func = struct2.get_polarizability_6x6

        # --- full coupling case
        # e,m <- e,m
        else:
            G_func = sim.environment.get_G_6x6
            alpha_func = struct2.get_polarizability_6x6
            selfterm_func = struct2.get_selfterm_6x6

        return G_func, alpha_func, selfterm_func

    # --- functions to setup interaction systems
    def _get_block_Gdotalpha(
        self,
        struct1: torch.Tensor,
        struct2: torch.Tensor,
        sim,
        wavelength: float,
    ) -> torch.Tensor:
        """calc. invertible interaction matrix $(G.alpha)$. selfterms on diagonal"""

        # process selfterms and polarizabilities
        G_func, alpha_func, selfterm_func = self.get_struct_pair_coupling_config(
            struct1, struct2, sim
        )

        # eval. Greens tensors between struct2 (emitter) and struct1 (observer)
        gdotalpha = G_func(
            struct1.positions.unsqueeze(1),  # light targets
            struct2.positions.unsqueeze(0),  # light sources
            wavelength=wavelength,
        )

        # if struct1 == struct2: put self-terms on diagonal (at R1[i]==R2[i])
        if struct1 == struct2:
            self_terms = selfterm_func(wavelength, sim.environment)

            N = len(struct1.positions)
            eye_idx = torch.arange(N, device=self.device)
            gdotalpha[eye_idx, eye_idx, :, :] = self_terms.unsqueeze(0)

        # multiply by polarizabilities
        polarizabilities = alpha_func(wavelength, sim.environment)
        gdotalpha = torch.matmul(gdotalpha, polarizabilities.unsqueeze(0))

        # reshape (N1,N2,6,6) to (6xN1, 6xN2) matrix (or (3,3), (6,3), (3,6) if not full coupling)
        gdotalpha_expanded = _reduce_dimensions(gdotalpha)

        return gdotalpha_expanded

    def get_interact(self, sim, wavelength: float, verbose=False) -> torch.Tensor:
        """get the full interaction block (E, H) of the linear system for a list of structures

        containes all direct and cross-contributions (E-E, H-H, E-H, H-E)

        Note: This implementation is not efficient for large numbers of structures.
        TODO: combine all structures of same interaction type before doing the simulation
        """
        from torchgdm.tools.misc import tqdm

        prgbar = True if verbose >= 3 else False

        # combine consecutive structures of same coupling type to reduce
        # memory transfer cost during iteration of all structures
        struct_list_reduced = [sim.structures[0]]
        if len(sim.structures) > 1:
            for _s in tqdm(
                sim.structures[1:], progress_bar=prgbar, title="optimizing structures"
            ):
                if _s.evaluation_terms == struct_list_reduced[-1].evaluation_terms:
                    try:
                        struct_list_reduced[-1] = struct_list_reduced[-1].combine(_s)
                    except:
                        # if combine fails, add as separate structure
                        struct_list_reduced.append(_s)
                else:
                    # if different coupling type, add as separate structure
                    struct_list_reduced.append(_s)

        if len(struct_list_reduced) > 10:
            warnings.warn(
                "Structure list could not be fully optimized. "
                + "For optimal performance, add structures with same "
                + "mesh or same coupling type in consecutive groups."
            )

        # calc. G dot alpha
        interact_rows = []
        for struct2 in tqdm(
            struct_list_reduced, progress_bar=prgbar, title="interaction matrix"
        ):
            interact_cols = []
            for struct1 in struct_list_reduced:
                # interaction block for struct2 (emitter) --> struct1 (observer)
                interact_expand = self._get_block_Gdotalpha(
                    struct1=struct1,
                    struct2=struct2,
                    sim=sim,
                    wavelength=wavelength,
                )
                interact_cols.append(interact_expand)
            interact_rows.append(torch.cat(interact_cols, dim=0))
        interact_full = torch.cat(interact_rows, dim=1)

        # free (GPU-)RAM
        purge_mem(interact_rows)

        # invertible matrix block is (1 - G.alpha): add the illumination field to output
        interact_full = (
            torch.eye(len(interact_full), device=self.device) - interact_full
        )

        return interact_full

    # --- field mask operations
    def get_all_fields_fill_masked_zeros(
        self,
        sim,
        field_at_polarizable: torch.Tensor,
        field_at_nonpolarizable: torch.Tensor = None,
    ) -> torch.Tensor:
        """create full field tensor with polarizable fields inserted.

        fields at non-polarizable positions can be optionally provided in a separate tensor.
        By default they are set to zero.

        Args:
            sim (`torchgdm.Simulation`): simulation instance
            field_at_polarizable (torch.Tensor): field values to be inserted at polarizable positions
            field_at_nonpolarizable (torch.Tensor, optional): field values to be inserted at non-polarizable positions. Defaults to None.

        Returns:
            torch.Tensor: full field tensor at all meshpoints with masked regions inserted
        """
        # init full internal field tensor
        fields_full = torch.zeros(
            (len(sim.illumination_fields), len(sim.get_all_positions()), 6),
            dtype=DTYPE_COMPLEX,
            device=self.device,
        )

        # get polarizable positions mask for field tensor
        mask_fields = sim._get_polarizable_mask_full_fields()

        # insert field elements at polarizable positions into masked full field tensor
        fields_full.flatten()[mask_fields.flatten()] = field_at_polarizable.flatten()

        # optional: inset field elements at non-polarizable positions into masked full field tensor
        if field_at_nonpolarizable is not None:
            fields_full.flatten()[
                ~mask_fields.flatten()
            ] = field_at_nonpolarizable.flatten()

        return fields_full

    def solve(
        self,
        sim,
        wavelength: float,
        batch_size=32,
        verbose: int = 1,
    ) -> torch.Tensor:
        """solve system of coupled structures for all illuminations

        Args:
            sim (`torchgdm.Simulation`): simulation instance
            wavelength (float): wavelength (nm)
            batch_size (int, optional): Nr of incident fields evaluated in parallel. Defaults to 32.
            verbose (int, optional): verbose level of status reporting. One of [0,1,2]. The higher, the more info are printed. Defaults to 1.

        Returns:
            torch.Tensor: list of internal fields, one for each illumination. Shape: (N_e0, nr of pos, 3)
        """
        if verbose >= 2:
            t_start = time.time()
            print("linsys", end="")

        # - setup interaction system
        interact = self.get_interact(sim, wavelength, verbose=verbose)
        if verbose >= 2:
            t_interact = time.time()
            print(" {:.2f}s. solve".format(t_interact - t_start), end="")

        # - inversion
        # LU, pivots, info = torch.linalg.lu_factor_ex(interact)  # experimental version, faster on GPU
        LU, pivots = torch.linalg.lu_factor(interact)
        if verbose >= 2:
            t_lu = time.time()
            print(
                " {:.2f}s. {}xE0".format(
                    t_lu - t_interact, len(sim.illumination_fields)
                ),
                end="",
            )

        # - evaluate zero order fields (illumination)
        f0 = sim._get_polarizablefields_e0_h0(wavelength)
        f0 = f0.view(len(f0), -1, 1)  # add dim for lu_solve

        # - solve for RHS vectors
        f_masked = _batched_lu_solve(LU, pivots, f0=f0, batch_size=batch_size)
        eh_inside = self.get_all_fields_fill_masked_zeros(sim, f_masked[..., 0])
        e_inside, h_inside = torch.chunk(eh_inside, 2, dim=2)  # separate E and H fields

        if verbose >= 2:
            t_solve = time.time()
            print(" {:.2f}s.".format(t_solve - t_lu), end="")

        return e_inside, h_inside

    def get_generalized_propagator(self, sim, wavelength: float) -> torch.Tensor:
        """solve system of coupled structures without illumination

        Args:
            sim (`torchgdm.Simulation`): simulation instance
            wavelength (float): wavelength (nm)

        Returns:
            torch.Tensor: full generalized propagator tensor for all polarizable elements
        """
        interact = self.get_interact(sim, wavelength)
        K = torch.linalg.inv(interact)
        return K



# encoding=utf-8
"""
2D visualization tools for vector fields
"""
# %%
import copy
import warnings

import torch

from torchgdm.constants import COLORS_DEFAULT
from torchgdm.tools.misc import to_np
from torchgdm.tools.misc import get_closest_wavelength
from torchgdm.visu.visu2d._tools import (
    _get_axis_existing_or_new,
    _get_closest_slice_level,
    _automatic_projection,
    _interpolate_to_grid,
    _apply_projection_to_vectorfield,
)


def vectorfield(field, illumination_index=0, whichfield="e", **kwargs):
    """quiver plot of 2D projection of 3D vectorsfield"""
    if whichfield.lower() == "e":
        f = field.efield[illumination_index]
    else:
        f = field.hfield[illumination_index]

    return _vectorfield(f, field.positions, **kwargs)


def streamlines_energy_flux(field, illumination_index=0, **kwargs):
    """plot energy flux streamlines"""
    f = field.get_energy_flux()[illumination_index]

    return _streamlines(f, field.positions, **kwargs)


def vectorfield_inside(sim, wavelength, illumination_index, whichfield="e", **kwargs):
    """quiver plot of 2D projection of 3D vectorfield inside structure"""
    wl = get_closest_wavelength(sim, wavelength)

    if len(sim.fields_inside) == 0:
        raise ValueError("No fields present. Run simulation first.")

    if illumination_index >= len(sim.fields_inside[wl].efield):
        raise ValueError(
            "Specified `illumination_index` out of range. Use an existing illumination."
        )

    if whichfield.lower() == "e":
        f = sim.fields_inside[wl].efield[illumination_index]
    else:
        f = sim.fields_inside[wl].hfield[illumination_index]

    return _vectorfield(f, sim.get_all_positions(), **kwargs)


def _vectorfield(
    field_vectors,
    positions,
    projection="auto",
    complex_part="real",
    slice_level=None,
    scale=10.0,
    sort_by_length=True,
    override_max_vector_length=False,
    cmap=None,
    set_ax_aspect=True,
    cmin=0.3,
    each_n_points=1,
    **kwargs,
):
    """quiver plot of 2D projection of 3D vectorsfield

    plot nearfield list as 2D vector plot, using matplotlib's `quiver`.
    `kwargs` are passed to `pyplot.quiver`

    Parameters
    ----------
    field_vectors : list of 3- or 6-tuples
        Nearfield definition. `np.array`, containing 6-tuples:
        (X,Y,Z, Ex,Ey,Ez), the field components being complex (use e.g.
        :func:`.tools.get_field_as_list`).
        Optionally can also take a list of 3-tuples (Ex, Ey, Ez),
        in which case the structure must be provided via the `struct` kwarg.

    positions : list or :class:`.core.simulation`, optional
        optional structure definition (necessary if field is supplied in
        3-tuple form without coordinates). Either `simulation` description
        object, or list of (x,y,z) coordinate tuples

    projection : str, default: 'auto'
        Which 2D projection to plot: "auto", "XY", "YZ", "XZ"

    complex_part : str, default: 'real'
        Which part of complex field to plot. Either 'real' or 'imag'.

    slice_level: float, default: `None`
        optional value of depth where to slice. eg if projection=='XY',
        slice_level=10 will take only values where Z==10.
            - slice_level = `None`, plot all vectors one above another without slicing.
            - slice_level = -9999 : take minimum value in field-list.
            - slice_level = 9999 : take maximum value in field-list.

    scale : float, default: 10.0
        optional vector length scaling parameter

    cmap : matplotlib colormap, default: `cm.Blues`
        matplotlib colormap to use for arrows (color scaling by vector length)

    set_ax_aspect : bool, default: True
        set aspect of matplotlib axes to "equal"

    cmin : float, default: 0.3
        minimal color to use from cmap to avoid pure white

    each_n_points : int, default: 1 [=all]
        show each N points only

    sort_by_length : bool, default: True
        sort vectors by length to avoid clipping (True: Plot longest
        vectors on top)

    override_max_vector_length : bool, default: False
        if True, use 'scale' as absolute scaling. Otherwise, normalize to
        field-vector amplitude.

    Returns
    -------

    return value of matplotlib's `quiver`

    """
    import matplotlib.pyplot as plt
    import matplotlib.cm as cm
    import matplotlib.colors as mcolors
    import numpy as np

    # - prep matplotlib, axes
    ax, show = _get_axis_existing_or_new()
    if cmap is None:
        cmap = cm.Blues
    if set_ax_aspect:
        ax.set_aspect("equal")

    # - to numpy, real part of pos
    _vec = to_np(field_vectors[::each_n_points])
    _pos = to_np(positions[::each_n_points]).real

    # - project on plane
    if projection.lower() == "auto":
        projection = _automatic_projection(_pos)

    vec, pos, levels = _apply_projection_to_vectorfield(
        _vec, _pos, projection, complex_part
    )

    # - optional slicing
    if slice_level is not None:
        slice_level = _get_closest_slice_level(levels, slice_level)
        vec = vec[(levels == slice_level)]
        pos = pos[(levels == slice_level)]

    # - sort and scale by length
    l_vec = np.linalg.norm(vec, axis=1)
    if sort_by_length:
        sort_idx = np.argsort(l_vec)
        pos = pos[sort_idx]
        vec = vec[sort_idx]
        l_vec = l_vec[sort_idx]

    if not override_max_vector_length:
        scale = scale * l_vec.max()
    else:
        scale = scale

    # -- quiver plot
    cscale = mcolors.Normalize(
        l_vec.min() - cmin * (l_vec.max() - l_vec.min()), l_vec.max()
    )  # colors by vector length

    im = ax.quiver(*pos.T, *vec.T, scale=scale, color=cmap(cscale(l_vec)), **kwargs)

    if show:
        plt.show()

    return im


def _streamlines(
    field_vectors,
    positions,
    projection="auto",
    complex_part="real",
    slice_level=None,
    set_ax_aspect=True,
    each_n_points=1,
    density=1,
    start_points="auto",
    broken_streamlines=False,
    **kwargs,
):
    """
    additional kwargs are passed to matplotlib's `streamplot`

    start_points: str
        - 'auto', use matplotlib default.
        - 'top': streamlines calc. starts from top
        - 'bottom', likewise but starting from bottom.
    """
    import matplotlib.pyplot as plt
    import matplotlib.cm as cm
    import matplotlib.colors as mcolors
    import numpy as np

    # - prep matplotlib, axes
    ax, show = _get_axis_existing_or_new()
    if set_ax_aspect:
        ax.set_aspect("equal")

    # - to numpy, real part of pos
    _vec = to_np(field_vectors[::each_n_points])
    _pos = to_np(positions[::each_n_points]).real

    # - project on plane
    if projection.lower() == "auto":
        projection = _automatic_projection(_pos)

    vec, pos, levels = _apply_projection_to_vectorfield(
        _vec, _pos, projection, complex_part
    )

    # - optional slicing
    if slice_level is not None:
        slice_level = _get_closest_slice_level(levels, slice_level)
        vec = vec[(levels == slice_level)]
        pos = pos[(levels == slice_level)]

    # - prepare the data: matplotlib's `streamplot` requires griddata
    v_x, extent, grid_x, grid_z = _interpolate_to_grid(
        pos[:, 0], pos[:, 1], vec[:, 0], return_grid=True
    )
    v_z, extent, grid_x, grid_z = _interpolate_to_grid(
        pos[:, 0], pos[:, 1], vec[:, 1], return_grid=True
    )

    # - optionally: start streamlines specifically from top / bottom
    N_pt_start = density * grid_x.shape[0]
    if start_points.lower() == "top":
        kwargs["start_points"] = np.stack(
            [
                np.linspace(
                    pos[:, 0].min(),
                    pos[:, 0].max(),
                    N_pt_start,
                ),
                np.ones(N_pt_start) * pos[:, 1].max(),
            ]
        ).T
    elif start_points.lower() == "bottom":
        kwargs["start_points"] = np.stack(
            [
                np.linspace(
                    pos[:, 0].min(),
                    pos[:, 0].max(),
                    N_pt_start,
                ),
                np.ones(N_pt_start) * pos[:, 1].min(),
            ]
        ).T

    # - plot
    stplt = plt.streamplot(
        grid_x,
        grid_z,
        v_x[::-1],
        v_z[::-1],
        density=density,
        broken_streamlines=broken_streamlines,
        **kwargs,
    )

    if show:
        plt.show()

    return stplt



# encoding=utf-8
"""
3D visualization tools for vector fields
"""
# %%
import warnings

import torch

from torchgdm.constants import COLORS_DEFAULT
from torchgdm.tools.misc import to_np
from torchgdm.tools.misc import get_closest_wavelength
from torchgdm.visu.visu2d._tools import (
    _get_axis_existing_or_new,
    _get_closest_slice_level,
)


def vectorfield(field, illumination_index=0, whichfield="e", **kwargs):
    """plot 3d vectorfield
    
    For doc, see :func:`_vectorfield`
    """
    if whichfield.lower() == "e":
        f = field.efield[illumination_index]
    else:
        f = field.hfield[illumination_index]

    return _vectorfield(f, field.positions, **kwargs)


def vectorfield_inside(sim, wavelength, illumination_index, whichfield="e", **kwargs):
    """plot 3d vectorfield inside structure
    
    For doc, see :func:`_vectorfield`
    """
    wl = get_closest_wavelength(sim, wavelength)

    if len(sim.fields_inside) == 0:
        raise ValueError("No fields present. Run simulation first.")

    if illumination_index >= len(sim.fields_inside[wl].efield):
        raise ValueError(
            "Specified `illumination_index` out of range. Use an existing illumination."
        )

    if whichfield.lower() == "e":
        f = sim.fields_inside[wl].efield[illumination_index]
    else:
        f = sim.fields_inside[wl].hfield[illumination_index]

    return _vectorfield(f, sim.get_all_positions(), **kwargs)


def _vectorfield(
    field_vectors,
    positions,
    complex_part="real",
    scale=10.0,
    override_max_vector_length=False,
    cmap="Blues",
    alpha=1,
    show_scalar_bar=False,
    each_n_points=1,
    pl=None,
    show="auto",
    **kwargs,
):
    """plot 2D Vector field as quiver plot

    plot nearfield list as 2D vector plot, using matplotlib's `quiver`.
    `kwargs` are passed to `pyVista.add_mesh`

    Parameters
    ----------
    field_vectors : list of 3- or 6-tuples
        Nearfield definition. `np.array`, containing 6-tuples:
        (X,Y,Z, Ex,Ey,Ez), the field components being complex (use e.g.
        :func:`.tools.get_field_as_list`).
        Optionally can also take a list of 3-tuples (Ex, Ey, Ez),
        in which case the structure must be provided via the `struct` kwarg.

    positions : list or :class:`.core.simulation`, optional
        optional structure definition (necessary if field is supplied in
        3-tuple form without coordinates). Either `simulation` description
        object, or list of (x,y,z) coordinate tuples

    complex_part : str, default: 'real'
        Which part of complex field to plot. Either 'real' or 'imag'.

    scale : float, default: 10.0
        optional vector length scaling parameter

    cmap : matplotlib colormap, default: `cm.Blues`
        matplotlib colormap to use for arrows (color scaling by vector length)

    alpha : float, default: 1
        opacity of vector arrows

    each_n_points : int, default: 1 [=all]
        show each N points only

    pl : plotter, default: None
        optional pre-existing pyvista plotter to plot into (for composition of scene).

    Returns
    -------

    return value of matplotlib's `quiver`

    """
    import pyvista as pv
    import numpy as np

    if "projection" in kwargs:
        kwargs.pop("projection")

    # - prep matplotlib, axes
    if type(show) == str:
        if show.lower() == "auto" and pl is None:
            show = True
        else:
            show = False
    if pl is None:
        pl = pv.Plotter()

    # - to numpy, take real or imag part
    f = to_np(field_vectors[::each_n_points])
    p = to_np(positions).real

    if complex_part.lower() == "real":
        f = f.real
    elif complex_part.lower() == "imag":
        f = f.imag
    else:
        raise ValueError(
            "Error: Unknown `complex_part` argument. Must be either 'real' or 'imag'."
        )

    # - vector length for color coding
    l_vec = np.linalg.norm(f, axis=1)

    if not override_max_vector_length:
        scale = scale * l_vec.max()
    else:
        scale = scale

    # - create pyvista data
    pts = pv.PolyData(p)
    pts["length"] = l_vec
    pts.set_active_scalars("length")
    pts["vectors"] = f * scale
    pts.set_active_vectors("vectors")

    # - do the actual plot
    pl.add_mesh(
        pts.arrows,
        cmap=cmap,
        show_edges=False,
        line_width=0.5,
        edge_opacity=alpha,
        opacity=alpha,
        show_scalar_bar=show_scalar_bar,
        name="vectorfield",
        **kwargs,
    )

    if show:
        pl.show()
        pl.camera.azimuth = 180.0  # convention: x-axis to the right

    return pts



# encoding=utf-8
# %%
import unittest

import torch
from scipy.special import jv, yv, iv, kv, hankel1, hankel2

import torchgdm as tg
from torchgdm.tools import special
from torchgdm.constants import DTYPE_FLOAT
from torchgdm.constants import DTYPE_COMPLEX


class TestSpecialFunctions(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing plane wave illumination.")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        self.z_re = torch.linspace(0.01, 5, 100, dtype=torch.float32)
        self.z_im = 1j * torch.linspace(0.01, 5, 100, dtype=torch.complex64)

    def test_bessel(self):
        for device in self.devices:
            for n in torch.arange(-3, 4):
                z_re = self.z_re.to(device)
                z_im = self.z_im.to(device)

                # compare with scipy - real args
                j_scipy = torch.as_tensor(
                    jv(n, tg.to_np(self.z_re)), device=device, dtype=DTYPE_FLOAT
                )
                j_torch = special.Jn(n, z_re)
                y_scipy = torch.as_tensor(
                    yv(n, tg.to_np(self.z_re)), device=device, dtype=DTYPE_FLOAT
                )
                y_torch = special.Yn(n, z_re)

                torch.testing.assert_close(j_scipy, j_torch, rtol=1e-3, atol=1e-3)
                torch.testing.assert_close(y_scipy, y_torch, rtol=1e-3, atol=1e-3)

                # compare with scipy - complex args
                j_scipy_im = torch.as_tensor(
                    jv(n, tg.to_np(self.z_im)), device=device, dtype=DTYPE_COMPLEX
                )
                j_torch_im = special.Jn(n, z_im)
                y_scipy_im = torch.as_tensor(
                    yv(n, tg.to_np(self.z_im)), device=device, dtype=DTYPE_COMPLEX
                )
                y_torch_im = special.Yn(n, z_im)

                torch.testing.assert_close(j_scipy_im, j_torch_im, rtol=1e-3, atol=1e-3)
                torch.testing.assert_close(y_scipy_im, y_torch_im, rtol=1e-3, atol=1e-3)

            if self.verbose:
                print("  - {}: field evaluation test passed.".format(device))

    def test_modbessel(self):
        for device in self.devices:
            for n in torch.arange(-3, 4):
                z_re = self.z_re.to(device)

                # modified bessel - real arg
                i_scipy = torch.as_tensor(
                    iv(n, tg.to_np(self.z_re)), device=device, dtype=DTYPE_FLOAT
                )
                i_torch = special.In(n, z_re)
                k_scipy = torch.as_tensor(
                    kv(n, tg.to_np(self.z_re)), device=device, dtype=DTYPE_FLOAT
                )
                k_torch = special.Kn(n, z_re)

                torch.testing.assert_close(i_scipy, i_torch, rtol=1e-3, atol=1e-3)
                torch.testing.assert_close(k_scipy, k_torch, rtol=1e-3, atol=1e-3)

    def test_hankel(self):
        for device in self.devices:
            for n in torch.arange(-3, 4):
                z_re = self.z_re.to(device)
                z_im = self.z_im.to(device)

                # hankel - real arg
                h1_scipy = torch.as_tensor(
                    hankel1(n, tg.to_np(self.z_re)), device=device, dtype=DTYPE_COMPLEX
                )
                h1_torch = special.H1n(n, z_re)
                h2_scipy = torch.as_tensor(
                    hankel2(n, tg.to_np(self.z_re)), device=device, dtype=DTYPE_COMPLEX
                )
                h2_torch = special.H2n(n, z_re)

                torch.testing.assert_close(h1_scipy, h1_torch, rtol=1e-3, atol=1e-3)
                torch.testing.assert_close(h2_scipy, h2_torch, rtol=1e-3, atol=1e-3)

                # hankel - imag arg
                h1_scipy_im = torch.as_tensor(
                    hankel1(n, tg.to_np(self.z_im)), device=device, dtype=DTYPE_COMPLEX
                )
                h1_torch_im = special.H1n(n, z_im)
                h2_scipy_im = torch.as_tensor(
                    hankel2(n, tg.to_np(self.z_im)), device=device, dtype=DTYPE_COMPLEX
                )
                h2_torch_im = special.H2n(n, z_im)

                torch.testing.assert_close(
                    h1_scipy_im, h1_torch_im, rtol=1e-3, atol=1e-3
                )
                torch.testing.assert_close(
                    h2_scipy_im, h2_torch_im, rtol=1e-3, atol=1e-3
                )


# %%


if __name__ == "__main__":
    print("testing special functions.")
    torch.set_printoptions(precision=7)
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# -*- coding: utf-8 -*-
"""package for various sorts of tools for torchgdm


Tools modules
-------------

.. currentmodule:: torchgdm.tools

.. autosummary::
   :toctree: generated/
    
    geometry
    batch
    misc
    interp
    special
    mie


Most relevant tools
-------------------

Many tools are used mainly internally. The most relevant tools are:


Geometry tools
^^^^^^^^^^^^^^

1D evaluation position generation:

.. autosummary::
   :toctree: generated/

   geometry.coordinate_map_1d
   geometry.coordinate_map_1d_circular
   geometry.coordinate_map_1d_circular_upper
   geometry.coordinate_map_1d_circular_lower
   geometry.sample_random_circular


2D evaluation position generation:

.. autosummary::
   :toctree: generated/

   geometry.coordinate_map_2d
   geometry.coordinate_map_2d_square
   geometry.coordinate_map_2d_spherical
   geometry.coordinate_map_2d_spherical_upper
   geometry.coordinate_map_2d_spherical_lower
   geometry.sample_random_spherical


Mie tools
^^^^^^^^^

.. autosummary::
   :toctree: generated/

   mie.mie_ab_cylinder_2d
   mie.mie_ab_sphere_3d
   mie.mie_crosssections_cylinder_2d
   mie.mie_crosssections_sphere_3d


"""
from . import geometry
from . import batch
from . import misc
from . import interp
from . import special
from . import mie



# -*- coding: utf-8 -*-
"""
various helper for torchgdm
"""
# %%
from typing import Optional
import warnings
import gc

import torch

try:
    from psutil import virtual_memory
except ModuleNotFoundError:

    def virtual_memory():
        warnings.warn("Package `psutil` not found. Install for automatic memory purge.")

        class mem_class:
            percent = 0.0

        return mem_class()


from torchgdm.constants import ERROR_ON_WAVELENGTH_MISSMATCH


# --- setting global device
def set_default_device(device):
    """set specific device as global default"""
    import torchgdm as tg
    import torchgdm.constants as tgc

    if str(device).lower().startswith("cuda"):
        if not torch.cuda.is_available():
            warnings.warn("CUDA not available. Fall back to 'cpu'.")
            device = "cpu"

    if str(device).lower().startswith("cpu"):
        if not torch.cpu.is_available():
            raise ValueError("Unexpected error. CPU not available.")

    tgc.DEFAULT_DEVICE = device
    tg.device = device


def get_default_device():
    """return current default device"""
    import torchgdm.constants as tgc

    return tgc.DEFAULT_DEVICE


def use_cuda(use_cuda: bool = True, verbose=True):
    """set first CUDA device as default"""
    if use_cuda:
        set_default_device("cuda")
    else:
        set_default_device("cpu")

    if verbose:
        import torchgdm.constants as tgc
        print("torchGDM: Default device set to '{}'".format(tgc.DEFAULT_DEVICE))
        if tgc.DEFAULT_DEVICE.lower().startswith("cuda"):
            print("          {}".format(torch.cuda.get_device_properties(tgc.DEFAULT_DEVICE).name))
            print("          total VRAM: {:.1f} MiB".format(torch.cuda.get_device_properties(tgc.DEFAULT_DEVICE).total_memory/1024**2))


# --- progress bar via tqdm
def tqdm(iterable, progress_bar=True, title=""):
    """wrapper to `tqdm` progress bar with optional `title`

    Args:
        iterable (iterable): iterable for loop over which to show a progress bar
        progress_bar (bool, optional): generic way to deactivate the progress bar. Defaults to True.
        title (str, optional): optional title of the progress bar. Defaults to "".

    Returns:
        iterable: wrapped iterable that will show a progress bar when iterated over
    """
    if progress_bar:
        try:
            from tqdm import tqdm

            pbar = tqdm(iterable)
            pbar.set_description(title)
            return pbar
        except (ModuleNotFoundError, ImportError):
            # if tqdm not installed, replace by nothing
            warnings.warn(
                "Progress bar requires `tqdm`, seems to be not installed. You may try 'pip install tqdm'."
            )
            return iterable
    else:
        return iterable


# # !! TODO
# class Memoize:
#     """decorator for memoizing"""
#     def __init__(self, func, len_memory=10000):
#         self.func = func
#         self.memory = {}
#         self.len_memory = len_memory
#         self.i_call = 0

#     def __call__(self, *args, **kwargs):
#         self.i_call += 1
#         if len(self.memory) > self.len_memory:
#             self.memory.clear()

#         # !!! TODO: Hashing of input arguments
#         hash_args = ...
#         if not hash_args in self.memory:
#             # print(self.i_call, len(self.memory), "not pre-calculated.")
#             self.memory[hash_args] = self.func(*args, **kwargs)
#         # else:
#             # print(self.i_call, len(self.memory), "already pre-calculated!")
#         return self.memory[hash_args]

#     def __get__(self, obj, objtype):
#         '''Support instance methods.'''
#         return functools.partial(self.__call__, obj)


def purge_mem(obj=None, dev=None, pruge_threshold=0.8):
    """purge RAM and GPU VRAM

    RAM will be purged only if occupied RAM > `pruge_threshold`

    Args:
        obj (any torch instance, optional): optional torch object to get device from. Defaults to None.
        dev (str or torch.device, optional): specify torch device for pruge. Defaults to None.
        pruge_threshold (float, optional): threshold above which to run garbage collector. Defaults to 0.8.
    """
    if obj is not None:
        try:
            dev = obj.device.type
        except AttributeError:
            dev = ""

        del obj

    # purge cuda cache if applicable
    if dev is not None:
        if type(dev) == torch.device:
            dev = dev.type
        if dev.startswith("cuda"):
            torch.cuda.empty_cache()

    # garbage collect if memory almost full
    if virtual_memory().percent > 0.8 * 100:
        gc.collect()


def test_positional_input(r1: torch.tensor, r2: torch.tensor):
    """test shapes of positional inputs for correct broadcasting"""
    if r1.device != r2.device:
        r2 = r2.to(r1.device)

    if len(r1.shape) == 1:
        if r1.shape[0] != 3:
            raise ValueError(
                "single point needs to be cartesian coordinate with 3 values."
            )
        # single point: expand dim
        r1 = r1[None, :]
        r2 = r2[None, :]
    elif len(r1.shape) == 2:
        if r1.shape[1] != 3:
            raise ValueError("points need to be cartesian coordinates with 3 values.")

    return r1, r2


def tensor_is_diagonal_and_identical(tensor: torch.Tensor):
    """Check if a tensor is diagonal and all elements are identical"""
    device = tensor.device
    # Check if all off-diagonal elements are zero
    off_diagonal_mask = ~torch.eye(tensor.size(0), dtype=torch.bool, device=device)
    if tensor.masked_select(off_diagonal_mask).nonzero().size(0) > 0:
        return False

    # Check if all diagonal elements are equal
    diag = tensor.diagonal()
    if not torch.all(diag == diag[0]):
        return False

    return True


def test_illumination_field_config_exists(
    sim,
    wavelength: float,
    illumination_index: int = None,
    raise_exception=True,
):
    """test if an illumination configuration exists and has been simulated yet

    Args:
        sim (:class:`torchgdm.Simulation`): simulation instance
        wavelength (float): in nm
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
        raise_exception (bool, optional): whether to raise an error if illumination not present or not evaluated. Defaults to True.

    Returns:
        bool: True if illumination exists and was calculated, else False
    """
    if float(wavelength) in sim.fields_inside:
        if illumination_index is not None:
            if (
                len(sim.fields_inside[float(wavelength)].efield) <= illumination_index
            ) or (illumination_index < 0):
                field_exists = False
            else:
                field_exists = True
        else:
            field_exists = True
    else:
        field_exists = False

    if len(sim.fields_inside) == 0:
        field_exists = False

    if field_exists:
        return True
    else:
        if raise_exception:
            raise ValueError(
                "Field configuration does not exist (wavelength: "
                + "{}nm, illumination index {}).".format(wavelength, illumination_index)
            )
        else:
            return False


def test_illumination_field_is_plane_wave(
    sim, illumination_index=None, raise_exception=False, message=""
):
    """test if the illumination is a plane wave

    This is used internally to verify correct usage of cross-section calculation

    Args:
        sim (:class:`torchgdm.Simulation`): simulation instance
        illumination_index (int, optional): optional index of a specific illumination. If None, batch-evaluate all illuminations. Defaults to None.
        raise_exception (bool, optional): If True, raise an error, if False just a warning. Defaults to False.
        message (str, optional): optional pre-posted message to warning / error. Defaults to "".

    Raises:
        Exception: if illumination is not a plane wave and `raise_exception=True`
    """
    pw_class_name = "free space plane wave"
    e_inc_names = [e_inc.__name__ for e_inc in sim.illumination_fields]
    if illumination_index is not None:
        e_inc_names = [e_inc_names[illumination_index]]

    for i, _n in enumerate(e_inc_names):
        if not _n.startswith(pw_class_name):
            if illumination_index is not None:
                i_f = illumination_index
            else:
                i_f = i

            if raise_exception:
                raise Exception(message + f"Field config #{i_f} is not a plane wave!")
            else:
                warnings.warn(message + f"Field config #{i_f} is not a plane wave!")


def get_closest_wavelength(sim, wavelength: float, raise_exception=None):
    """get closest available wavelength in the simulation

    Args:
        sim (:class:`torchgdm.Simulation`): simulation instance
        wavelength (float): target wavelength in nm
        raise_exception (bool, optional): If None (default), use global configuration in `torchgdm.constants.ERROR_ON_WAVELENGTH_MISSMATCH`. If True, raise an error if exact wavelength not available, if False just emit a warning. Defaults to None.

    Raises:
        ValueError: if wavelength is not available and `raise_exception=True`

    Returns:
        float: closest available match to `wavelength`
    """
    wavelength = float(wavelength)

    if wavelength not in sim.fields_inside:
        if ERROR_ON_WAVELENGTH_MISSMATCH:
            raise ValueError(
                "Requested wavelength ({}nm) not available.".format(wavelength)
            )
        else:
            wl = float(
                sim.wavelengths[torch.argmin(torch.abs(wavelength - sim.wavelengths))]
            )
            warnings.warn(
                "Requested wavelength ({}nm) not available. Use closest match: {}nm.".format(
                    wavelength, wl
                )
            )
    else:
        wl = wavelength

    return wl


def ptp(input: torch.Tensor, dim: Optional[int] = None, keepdim=False) -> torch.Tensor:
    """Range of values (maximum - minimum) along an axis.

    simple torch implementation of :func:`numpy.ptp`

    Args:
        input (torch.Tensor): Input values.
        dim (Optional[int], optional): Axis along which to find the peaks. Defaults to None.
        keepdim (bool, optional): If this is set to True, the axes which are reduced are left in the result as dimensions with size one. Defaults to False.

    Returns:
        torch.Tensor: The range of a given input tensor - scalar if one-dimensional or an array holding the result along the given axis
    """
    if dim is None:
        return input.max() - input.min()
    return input.max(dim, keepdim).values - input.min(dim, keepdim).values


def to_np(input: torch.Tensor):
    """detach tensor and convert to numpy array

    Args:
        input (torch.Tensor): tensor to convert to numpy

    Returns:
        numpy.ndarray: input as numpy array
    """
    try:
        return input.detach().cpu().numpy()
    except AttributeError:
        warnings.warn("Torch.Tensor detach failed. Returning original input instead.")
        return input



# encoding=utf-8
"""
2D visualization tools for geometries

"""
# %%
import copy
import itertools
import warnings

import torch

from torchgdm.constants import COLORS_DEFAULT
from torchgdm.tools.misc import to_np
from torchgdm.tools.geometry import get_projection, get_step_from_geometry
from torchgdm.visu.visu2d._tools import (
    _get_axis_existing_or_new,
    _automatic_projection,
    _optimize_alpha_value,
)


# global color handling
COLOR_ITERATOR = itertools.cycle(COLORS_DEFAULT)
LEGEND_ENTRIES_LOOKUP = dict()


# reset color iterator and materials lookup
def _reset_color_iterator():
    global COLOR_ITERATOR, LEGEND_ENTRIES_LOOKUP
    COLOR_ITERATOR = itertools.cycle(COLORS_DEFAULT)
    LEGEND_ENTRIES_LOOKUP = dict()


def _return_next_color():
    return next(COLOR_ITERATOR)


_reset_color_iterator()


def _generate_legend(ax):
    import matplotlib.pyplot as plt

    global LEGEND_ENTRIES_LOOKUP

    if len(LEGEND_ENTRIES_LOOKUP) != 0:
        labels = LEGEND_ENTRIES_LOOKUP.keys()
    else:
        labels = []

    for i_s, label in enumerate(labels):
        legend_dict = LEGEND_ENTRIES_LOOKUP[label]
        plt.plot(
            [],
            [],
            marker=legend_dict["marker"],
            lw=0,
            markerfacecolor=legend_dict["fc"],
            markeredgecolor=legend_dict["ec"],
            markersize=legend_dict["markersize"],
            label=label,
        )

    return ax.legend()


def _plot_structure_discretized(
    struct,
    projection="auto",
    scale=1.0,
    color="auto",
    show_grid=True,
    legend=False,
    set_ax_aspect=True,
    alpha=1.0,
    reset_color_cycle=True,
    **kwargs,
):
    import matplotlib
    import matplotlib.pyplot as plt
    import numpy as np

    global COLOR_ITERATOR, LEGEND_ENTRIES_LOOKUP

    if reset_color_cycle:
        _reset_color_iterator()

    # prep
    grid_kwargs = dict(ec="k", lw=0.5) if show_grid else dict()
    ax, show = _get_axis_existing_or_new()
    if set_ax_aspect:
        ax.set_aspect("equal")

    # get geometry projection
    if projection.lower() == "auto":
        projection = _automatic_projection(struct.positions)
    pos_proj, idx_proj = get_projection(
        struct.positions, projection, return_inverse=True
    )

    if struct.step.dim() == 0:
        steplist = torch.tile(struct.step, (len(struct.positions),))
    else:
        steplist = struct.step
    steplist_proj = steplist[idx_proj]

    # convert to numpy
    pos_proj = to_np(pos_proj)
    idx_proj = to_np(idx_proj)
    steplist_proj = to_np(steplist_proj)

    # multi-material: different colors for parts of same material
    if color == "auto":
        diff_mat_names = [s.__name__ for s in struct.materials]
        mat_names = np.array(diff_mat_names)[idx_proj]

        different_materials = np.unique(mat_names)

        mat_pos_subset_idx = []
        for pos_single_mat in different_materials:
            mat_pos_subset_idx.append(
                np.arange(len(mat_names))[mat_names == pos_single_mat]
            )

    else:
        different_materials = ["struct. id:{}".format(struct.id)]
        mat_pos_subset_idx = [np.arange(len(pos_proj))]

    # the actual plot: create square patches for each meshcell
    # colors_for_legend = []
    for i_s, pos_idx in enumerate(mat_pos_subset_idx):
        x, y = pos_proj[pos_idx].T[:2]
        steplist_sub = steplist_proj[pos_idx]
        patches = []

        # rectangles for each mesh-cell
        for i_pt, [xi, yi] in enumerate(zip(x, y)):
            step = scale * steplist_sub[i_pt]
            step_plt = float(step / struct.mesh_normalization_factor)
            xis = xi - step_plt / 2
            yis = yi - step_plt / 2

            # TODO: plot actual hex mesh as hexagons
            patches.append(
                plt.Rectangle((xis, yis), width=step_plt, height=step_plt, linewidth=0)
            )

        # chose color
        if color == "auto":
            # if material already used, re-use its color. otherwise use next color
            if different_materials[i_s] in LEGEND_ENTRIES_LOOKUP:
                _col = LEGEND_ENTRIES_LOOKUP[different_materials[i_s]]["color"]
            else:
                _col = next(COLOR_ITERATOR)
        elif type(color) in [list, tuple]:
            _col = color[i_s]
        else:
            _col = color

        # add legend entry:
        LEGEND_ENTRIES_LOOKUP[different_materials[i_s]] = dict(
            color=_col,
            fc=_col,
            ec=_col,
            marker="s",
            markersize=10,
        )

        # add patch collection to plot
        kwargs.pop("color_circle", None)
        c = matplotlib.collections.PatchCollection(
            patches, fc=_col, alpha=alpha, **grid_kwargs, **kwargs
        )
        ax.add_collection(c)

    # optional plot legend
    if legend:
        _generate_legend(ax)

    # config axes
    plt.autoscale(tight=False)
    plt.xlabel("{} (nm)".format(projection[0]))
    plt.ylabel("{} (nm)".format(projection[1]))

    if show:
        plt.show()
    else:
        return ax


def _plot_contour_discretized(
    struct,
    projection="auto",
    color="auto",
    set_ax_aspect=True,
    alpha=1.0,
    alpha_value=None,
    **kwargs,
):
    """plot the concave hull around 2D projection of a structure discretization

    Args:
        struct (struct or torch.Tensor): structure or torch.Tensor of discretized object
        projection (str, optional): Which 2D projection plane to use. Defaults to "auto".
        color (str, optional): optional matplotlib compatible color. Defaults to "auto".
        set_ax_aspect (bool, optional): If True, will set aspect of plot to equal. Defaults to True.
        alpha (float, optional): matplotlib transparency value. Defaults to 1.0.
        alpha_value (str or float, optional): alpha_value for concave hull calculation. If None, use 1/step. If "opt", or "optimize", use a simple optimization scheme. Defaults to None.

    Raises:
        ValueError: no solution found

    Returns:
        matplotlib axes
    """
    import matplotlib.pyplot as plt
    import alphashape
    import logging

    global COLOR_ITERATOR, LEGEND_ENTRIES_LOOKUP

    # prep
    ax, show = _get_axis_existing_or_new()
    if set_ax_aspect:
        ax.set_aspect("equal")

    # automatic next color
    if color == "auto":
        color = next(COLOR_ITERATOR)

    if type(struct) == torch.Tensor:
        positions = struct.to("cpu")
        mesh_norm_factor = 1.0
    else:
        positions = struct.positions.to("cpu")
        mesh_norm_factor = struct.mesh_normalization_factor.to("cpu")

    # get geometry projection
    if projection.lower() == "auto":
        projection = _automatic_projection(positions)
    pos_proj, idx_proj = get_projection(positions, projection, return_inverse=True)

    # expand geometry by a half-step (outer points)
    step = get_step_from_geometry(pos_proj)
    step_plt = step / mesh_norm_factor
    pos_px = pos_proj + torch.as_tensor([[step_plt / 2, step_plt / 2]])
    pos_mx = pos_proj + torch.as_tensor([[-step_plt / 2, step_plt / 2]])
    pos_py = pos_proj + torch.as_tensor([[step_plt / 2, -step_plt / 2]])
    pos_my = pos_proj + torch.as_tensor([[-step_plt / 2, -step_plt / 2]])
    pos2d_exp = torch.concatenate([pos_px, pos_mx, pos_py, pos_my])

    # concave hull (alpha shape) - ignore warnings
    if alpha_value is None:
        alpha_value = 1.0 / step
    if type(alpha_value) == str:
        if alpha_value.lower() in ["opt", "optimize"]:
            alpha_value = _optimize_alpha_value(pos2d_exp)

    original_log_level = logging.getLogger().getEffectiveLevel()
    logging.disable(logging.WARN)
    geo_alpha = alphashape.alphashape(pos2d_exp, alpha_value)
    logging.disable(original_log_level)
    # in future: maybe replace by a simple local implementation to avoid dependencies
    # edges = _alpha_shape(pos2d_exp, alpha_value)

    if geo_alpha.geom_type == "Polygon":
        g_a_list = [geo_alpha]
    elif geo_alpha.geom_type == "MultiPolygon":
        g_a_list = [_g for _g in geo_alpha.geoms]
    else:
        raise ValueError("Unexpected error: Unknown alphashape geometry.")

    for _geo_a in g_a_list:
        # plot every sub geometry
        xx, yy = _geo_a.exterior.coords.xy
        xx, yy = xx.tolist(), yy.tolist()
        ax.plot(xx, yy, color=color, alpha=alpha, **kwargs)

    ax.autoscale(tight=False)
    ax.set_xlabel("{} (nm)".format(projection[0]))
    ax.set_ylabel("{} (nm)".format(projection[1]))

    if show:
        plt.show()

    return ax


def _plot_structure_eff_pola(
    struct,
    projection="auto",
    scale=1.0,
    color="auto",
    linestyle_circle=(0, (2, 2)),
    color_circle="auto",
    color_circle_fill=None,
    alpha=1,
    show_grid=True,
    color_grid="auto",
    alpha_grid=0.25,
    legend=False,
    set_ax_aspect=True,
    reset_color_cycle=True,
    **kwargs,
):
    import matplotlib
    import matplotlib.pyplot as plt

    global COLOR_ITERATOR, LEGEND_ENTRIES_LOOKUP

    if reset_color_cycle:
        _reset_color_iterator()

    ax, show = _get_axis_existing_or_new()
    if set_ax_aspect:
        ax.set_aspect("equal")

    # legend label based on structure ID
    legend_label = "eff.pola. (id: {})".format(struct.id)

    # automatic next color
    if color == "auto":
        if legend_label in LEGEND_ENTRIES_LOOKUP:
            _col = LEGEND_ENTRIES_LOOKUP[legend_label]["color"]
        else:
            _col = next(COLOR_ITERATOR)
    else:
        _col = color

    if color_grid == "auto":
        color_grid = _col
    if color_circle == "auto":
        color_circle = _col
    if color_circle_fill is None:
        color_circle_fill = (1, 1, 1, 0.0)

    # add / update legend entry
    LEGEND_ENTRIES_LOOKUP[legend_label] = dict(
        color=_col,
        fc=color_circle_fill,
        ec=_col,
        marker="o",
        markersize=10,
    )

    # get geometry projection
    if projection.lower() == "auto":
        projection = _automatic_projection(struct.positions)
    pos_proj = get_projection(struct.positions, projection, unique=False)

    if struct.step.dim() == 0:
        steplist = torch.tile(struct.step, (len(struct.positions),))
    else:
        steplist = struct.step

    # convert to numpy
    pos_proj = to_np(pos_proj)
    steplist = to_np(steplist)

    # plot enclosing spheres for eff. polas. structures
    patches = []
    for i_s, [xi, yi] in enumerate(zip(*pos_proj.T)):
        patches.append(
            plt.Circle((xi, yi), radius=steplist[i_s] / 2.0, fc=color_circle_fill)
        )
    c = matplotlib.collections.PatchCollection(
        patches,
        fc=color_circle_fill,
        ec=color_circle,
        linestyle=linestyle_circle,
        linewidth=1,
    )
    ax.add_collection(c)

    # optionally, plot all points of represented geometries
    if show_grid:
        for subgeo in struct.full_geometries:
            geo_step = to_np(get_step_from_geometry(torch.as_tensor(subgeo)))
            subgeo_proj, idx_proj = get_projection(
                subgeo, projection, return_inverse=True
            )
            Xsg, Ysg = to_np(subgeo_proj).T[:2]

            patches2 = []
            for i_pt, [xi, yi] in enumerate(zip(Xsg, Ysg)):
                step = scale * geo_step
                xis = xi - step / 2
                yis = yi - step / 2
                patches2.append(plt.Rectangle((xis, yis), width=step, height=step))
            c2 = matplotlib.collections.PatchCollection(
                patches2,
                color=color_grid,
                ec="k",
                lw=0.5,
                alpha=alpha * alpha_grid,
                **kwargs,
            )
            ax.add_collection(c2)

    # optional legend additions
    if legend:
        _generate_legend(ax)

    # plot centers of mass
    plt.scatter(*pos_proj.T, s=scale * 25, marker="x", c="k", alpha=alpha, **kwargs)

    # config axes
    plt.autoscale(tight=False)
    plt.xlabel("{} (nm)".format(projection[0]))
    plt.ylabel("{} (nm)".format(projection[1]))

    return ax


def structure(
    struct,
    projection="auto",
    color="auto",
    legend=True,
    reset_color_cycle=True,
    **kwargs,
):
    """plot 2D projection of structure

    plot the structure `struct` as a scatter projection to a carthesian plane.
    Either from a structure instance, or using a simulation as input.

    kwargs are passed to individual structure plotting and / or to matplotlib's `scatter`

    Parameters
    ----------
    struct : simulation or structure
          either a simulation or a structure instance

    projection : str, default: 'auto'
        which 2D projection to plot: "auto", "XY", "YZ", "XZ"

    color : str or matplotlib color, default: "auto"
            Color of scatterplot. Either "auto", or matplotlib-compatible color.
            "auto": automatic color selection (multi-color for multiple materials).

    legend : bool, default: True
        whether to add a legend if multi-material structure (requires auto-color enabled)


    Returns
    -------
    result returned by matplotlib's `scatter`

    """
    import matplotlib.pyplot as plt
    from torchgdm.simulation import SimulationBase, Simulation
    from torchgdm.struct.volume.pola import StructDiscretized3D
    from torchgdm.struct.volume.pola import StructDiscretizedCubic3D
    from torchgdm.struct.volume.pola import StructDiscretizedHexagonal3D
    from torchgdm.struct.point.pola import StructEffPola3D

    if reset_color_cycle:
        _reset_color_iterator()

    # got a structure instance:
    if (
        issubclass(type(struct), StructDiscretized3D)
        or issubclass(type(struct), StructDiscretizedCubic3D)
        or issubclass(type(struct), StructDiscretizedHexagonal3D)
    ):
        im = _plot_structure_discretized(
            struct,
            projection=projection,
            color=color,
            legend=legend,
            reset_color_cycle=False,
            **kwargs,
        )

    # got a structure effective pola. instance
    elif issubclass(type(struct), StructEffPola3D):
        im = _plot_structure_eff_pola(
            struct,
            projection=projection,
            color=color,
            legend=legend,
            reset_color_cycle=False,
            **kwargs,
        )

    # got a simulation instance
    elif issubclass(type(struct), Simulation) or issubclass(
        type(struct), SimulationBase
    ):
        # -- prep
        sim = struct
        ax, show = _get_axis_existing_or_new()

        if projection.lower() == "auto":
            if len(sim.get_all_positions())>1:
                projection = _automatic_projection(sim.get_all_positions())
                

        # -- call each structure's plot functions
        for i_s, _st in enumerate(sim.structures):
            im = _st.plot(
                projection=projection,
                color=color,
                legend=False,
                reset_color_cycle=False,
                **kwargs,
            )

        # -- finalize: config global plot
        if legend:
            _generate_legend(ax)
        if show:
            plt.show()
    else:
        raise ValueError("Unknown structure input")

    if reset_color_cycle:
        _reset_color_iterator()

    return im


def contour(
    struct,
    projection="auto",
    color="auto",
    set_ax_aspect=True,
    alpha=1.0,
    alpha_value=None,
    reset_color_cycle=True,
    **kwargs,
):
    """plot contour of a structure's 2D projection

    plot the contour of structure `struct` as a scatter projection to a carthesian plane.
    Either from list of coordinates, structure object or a simulation instance.
    For effective polarizability structures, will try to use the full mesh if available.

    Note: Doesn't support legend entries so far.

    Args:
        struct (Simulation, Structure): either a simulation or a structure instance
        projection (str, optional): which cartesian plane to project onto. Defaults to "auto".
        color (str, optional): optional matplotlib compatible color. Defaults to "auto".
        set_ax_aspect (bool, optional): If True, will set aspect of plot to equal. Defaults to True.
        alpha (float, optional): matplotlib transparency value. Defaults to 1.0.
        alpha_value (float, optional): alphashape value. If `None`, try to automatically optimize for best enclosing of structure. Defaults to None.
        reset_color_cycle (bool, optional): reset color cycle after finishing the plot. Defaults to True.

    Raises:
        ValueError: Invalid structure input

    Returns:
        matplotlib line: matplotlib's `scatter` output
    """

    import matplotlib.pyplot as plt
    from torchgdm.simulation import SimulationBase, Simulation
    from torchgdm.struct.volume.pola import StructDiscretized3D
    from torchgdm.struct.volume.pola import StructDiscretizedCubic3D
    from torchgdm.struct.volume.pola import StructDiscretizedHexagonal3D
    from torchgdm.struct.point.pola import StructEffPola3D

    if reset_color_cycle:
        _reset_color_iterator()

    if (
        issubclass(type(struct), StructDiscretized3D)
        or issubclass(type(struct), StructDiscretizedCubic3D)
        or issubclass(type(struct), StructDiscretizedHexagonal3D)
        or type(struct) == torch.Tensor
    ):
        return _plot_contour_discretized(
            struct,
            projection=projection,
            color=color,
            alpha=alpha,
            alpha_value=alpha_value,
            set_ax_aspect=set_ax_aspect,
            **kwargs,
        )
    elif issubclass(type(struct), StructEffPola3D):
        if len(struct.full_geometries) == 0:
            warnings.warn("No mesh grid data available. Skipping.")
            return None
        for subgeo in struct.full_geometries:
            return _plot_contour_discretized(
                subgeo,
                projection=projection,
                color=color,
                alpha=alpha,
                alpha_value=alpha_value,
                set_ax_aspect=set_ax_aspect,
                **kwargs,
            )
    elif issubclass(type(struct), Simulation) or issubclass(
        type(struct), SimulationBase
    ):
        # -- prep
        sim = struct
        ax, show = _get_axis_existing_or_new()

        if projection.lower() == "auto":
            projection = _automatic_projection(sim.get_all_positions())

        # -- call each structure's plot functions
        for i_s, _st in enumerate(sim.structures):
            im = _st.plot_contour(
                projection=projection,
                color=color,
                alpha=alpha,
                alpha_value=alpha_value,
                **kwargs,
            )

        # -- finalize: config global plot
        if show:
            plt.show()

        return im
    else:
        raise ValueError("Unknown structure input")


if __name__ == "__main__":
    print("test")
    import time
    import torchgdm as tg
    import matplotlib.pyplot as plt

    # - full structure: volume discretization
    step = 25.0  # nm
    r_disc = 120 / step  # units of step
    h_disc = 2
    material = tg.materials.MatConstant(eps=12.0)
    t0 = time.time()
    geometry = tg.struct.volume.discretizer_hexagonalcompact(
        *tg.struct.volume.disc(r_disc, h_disc), step=step, orientation=1
    )
    struct = tg.struct.StructDiscretizedHexagonal3D(geometry, material)
    # geometry = tg.struct.volume.discretizer_cubic(
    #     *tg.struct.volume.disc(r_disc, h_disc), step=step
    # )
    # struct = tg.struct.StructDiscretizedCubic3D(geometry, material)

    print(struct)
    print("Time meshing".format(time.time() - t0))

    fig, ax = plt.subplots()
    structure(struct)
    _plot_contour_discretized(struct)
    plt.show()



# -*- coding: utf-8 -*-
"""2D effective line polarizabilities

.. currentmodule:: torchgdm.struct.line_2d

.. autosummary::
   :toctree: generated/
   
   StructEffPola2D
   StructMieCylinderEffPola2D

"""
from .pola import StructEffPola2D
from .pola import StructMieCylinderEffPola2D
from . import pola



# -*- coding: utf-8 -*-
"""package for torchgdm structures

Contains Classes for 3D and 2D discretized structures and point (3D) and line (2D) effective polarizability structures:

.. currentmodule:: torchgdm.struct

3D
--

.. autosummary::
   :toctree: generated/
   
   StructDiscretizedCubic3D
   StructDiscretizedHexagonal3D
   StructEffPola3D
   StructMieSphereEffPola3D


3D discretization
-----------------

.. autosummary::
   :toctree: generated/
   
   volume


2D
--

.. autosummary::
   :toctree: generated/
   
   StructDiscretizedSquare2D
   StructEffPola2D
   StructMieCylinderEffPola2D


2D discretization
-----------------

.. autosummary::
   :toctree: generated/
   
   surface_2d

base class
----------

.. autosummary::
   :toctree: generated/
   
   StructBase

"""
from . import volume
from . import point
from . import surface_2d
from . import line_2d

# - 3D
from .volume.pola import StructDiscretized3D
from .volume.pola import StructDiscretizedCubic3D
from .volume.pola import StructDiscretizedHexagonal3D

from .point.pola import StructEffPola3D
from .point.pola import StructMieSphereEffPola3D  # Mie core-shell sphere

# - 2D
from .surface_2d.pola import StructDiscretized2D
from .surface_2d.pola import StructDiscretizedSquare2D
from .line_2d.pola import StructMieCylinderEffPola2D  # Mie core-shell cylinder

from .line_2d.pola import StructEffPola2D

# - base class
from .base_classes import StructBase


# encoding=utf-8
"""2D homogeneous environment and illuminations

Classes
-------

along with the main environment definition, compatible illuminations are defined:

.. currentmodule:: torchgdm.env.freespace_2d

.. autosummary::
   :toctree: generated/

   EnvHomogeneous2D
   NullField
   PlaneWave
   ElectricLineDipole
   MagneticLineDipole

"""
from . import inc_fields

from .dyads import EnvHomogeneous2D

from .inc_fields import NullField
from .inc_fields import PlaneWave
from .inc_fields import ElectricLineDipole
from .inc_fields import MagneticLineDipole


# encoding=utf-8
#
# Copyright (C) 2023-2025, P. R. Wiecha
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
TorchGDM - nano-optics + auto-diff

A full-field electrodynamical Green's Dyadic Method solver with 
support for effective E/M polarizabilities, written in pytorch. 




Core classes
============

Two classes provide high-level access to the core functionalities:

.. currentmodule:: torchgdm

.. autosummary::
   :recursive:
   :toctree: generated/

   Simulation
   Field
   

Simulation description
======================

The simulation takes a list of structures, the environment and a list 
of illumination fields as main inputs. 
The respective classes are found in following sub packages:

Structures
----------

The sub-packages :mod:`struct.volume` and :mod:`struct.surface_2d` 
contain also discretizations for a selection of geometric primitives.

.. autosummary::
   :toctree: generated/

   struct
   struct.volume
   struct.point
   struct.surface_2d
   struct.line_2d


Materials
----------

.. autosummary::
   :toctree: generated/

   materials


Environments and illuminations
------------------------------

.. autosummary::
   :toctree: generated/

   env.freespace_2d
   env.freespace_3d

*Note:* Illumination fields are associated with each environment. 
They can be found in the respective base environment package
(:mod:`env.freespace_2d` and :mod:`env.freespace_3d`).


Postprocessing and Tools
========================

Most postprocessing and visualization tools are accessible also through 
the :class:`Simulation` and :class:`Field` classes. 
Alternatively the functional API can be used:

Post processing
---------------

.. autosummary::
   :toctree: generated/

   postproc
   postproc.crosssect
   postproc.fields
   postproc.multipole
   postproc.green


Visualization
-------------

.. autosummary::
   :toctree: generated/
   :recursive:

   visu.visu2d
   visu.visu3d


Tools
-----

For detailed documentation, see:

.. autosummary::
   :toctree: generated/

    tools


Many tools are used mainly internally. The most relevant tool modules are therefore:

.. autosummary::
   :toctree: generated/
    
    tools.geometry
    tools.mie


CUDA GPU acceleration
=====================

TorchGDM exposes a few top level helper functions for CUDA usage:

.. autosummary::
   :toctree: generated/

    to_np
    use_cuda
    set_default_device
    get_default_device


"""

__name__ = "torchgdm"
__version__ = "0.46"
__date__ = "02/01/2025"  # MM/DD/YYY
__license__ = "GPL3"
__status__ = "beta"

__copyright__ = "Copyright 2023-2025, Peter R. Wiecha"
__author__ = "Peter R. Wiecha"
__maintainer__ = "Peter R. Wiecha"
__email__ = "pwiecha@laas.fr"
__credits__ = [
    "Christian Girard",
    "Arnaud Arbouet",
    "Clément Majorel",
    "Sofia Ponomareva",
    "Antoine Azéma",
    "Adelin Patoux",
    "Renaud Marty",
]

# --- populate namespace
from .constants import DEFAULT_DEVICE

device = DEFAULT_DEVICE

# make some functions and classes available at top level
from .simulation import Simulation
from .field import Field
from .tools.misc import to_np, tqdm
from .tools.misc import use_cuda, get_default_device, set_default_device

# make some submodules available at top level
from .visu import visu2d
from .visu import visu3d

# modules
from . import constants
from . import linearsystem
from . import simulation
from . import field

# sub packages
from . import env
from . import materials
from . import struct
from . import postproc
from . import tools



# encoding=utf-8
"""
2D visualization tools for vector fields
"""
# %%
import copy
import warnings

import torch

from torchgdm.constants import COLORS_DEFAULT
from torchgdm.tools.misc import to_np
from torchgdm.visu.visu2d._tools import (
    _get_axis_existing_or_new,
    _get_closest_slice_level,
    _interpolate_to_grid,
    _automatic_projection,
    _apply_projection,
)


def field_intensity(field, illumination_index=0, whichfield="e", **kwargs):
    """2D colorplot of field intensity"""
    if whichfield.lower() == "e":
        f = field.get_efield_intensity()[illumination_index]
    else:
        f = field.get_hfield_intensity()[illumination_index]

    return _scalarfield(f, field.positions, **kwargs)


def field_amplitude(
    field,
    illumination_index=0,
    whichfield="e",
    complex_part="re",
    field_component="x",
    **kwargs,
):
    """2D colorplot of a field amplitude component"""
    if whichfield.lower() == "e":
        f = field.efield[illumination_index]
    else:
        f = field.hfield[illumination_index]

    if complex_part.lower() in ["re", "real"]:
        f = f.real
    elif complex_part.lower() in ["im", "imag"]:
        f = f.imag
    else:
        raise ValueError("Unknown complex part. Use either 're' or 'im'.")

    if field_component.lower() == "x":
        f = f[:, 0]
    elif field_component.lower() == "y":
        f = f[:, 1]
    elif field_component.lower() == "z":
        f = f[:, 2]
    else:
        raise ValueError("Unknown field_component. Use either 'x', 'y' or 'z'.")

    return _scalarfield(f, field.positions, **kwargs)


def _scalarfield(
    field_scalars,
    positions,
    projection="auto",
    slice_level=None,
    interpolation="linear",
    set_ax_aspect=True,
    **kwargs,
):
    """2D colorplot of a scalar field

    plot nearfield list as 2D vector plot, using matplotlib's `quiver`.
    `kwargs` are passed to `pyplot.quiver`

    Args:
        field_scalars (torch.Tensor): list of scalar values
        positions (torch.Tensor): list of (x,y,z) coordinates
        projection (str, optional): Which 2D projection to plot: "auto", "XY", "YZ", "XZ". Defaults to "auto".
        slice_level (float, optional): depth where to slice. eg if projection=='XY', slice_level=10 will take only values where Z==10. If no positions exist with the selected coordinate value, take the closest existing value. Defaults to None.
        interpolation (str, optional): interpolation method for scipy `grid_data`. Can be 'linear' or 'nearest'. See `scipy.interpolate.griddata` for details. Defaults to "linear".
        set_ax_aspect (bool, optional): set aspect of matplotlib axes to "equal". Defaults to True.

    Returns:
        output of matplotlib's `quiver`
    """
    import matplotlib.pyplot as plt
    import matplotlib.cm as cm
    import numpy as np

    if type(positions) == dict:
        positions = positions["r_probe"]

    assert len(field_scalars) == len(positions)

    # - to numpy, take real or imag part
    s = to_np(field_scalars).real
    _p = to_np(positions).real

    # - select projection
    if projection.lower() == "auto":
        projection = _automatic_projection(_p)
    p, levels = _apply_projection(_p, projection)

    # - optional slicing
    if slice_level is not None:
        slice_level = _get_closest_slice_level(levels, slice_level)
        p = p[(levels == slice_level)]
        s = s[(levels == slice_level)]

    # - interpolate grid data
    map_2d, extent = _interpolate_to_grid(
        p[:, 0], p[:, 1], s, interpolation=interpolation
    )

    # - prep matplotlib, axes
    ax, show = _get_axis_existing_or_new()
    if set_ax_aspect:
        ax.set_aspect("equal")

    # plot
    im = ax.imshow(map_2d, extent=extent, **kwargs)

    if show:
        plt.show()

    return im



# -*- coding: utf-8 -*-
"""
torch-based linear interpolation
"""
# %%
from itertools import product

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device


class RegularGridInterpolator:
    """
    `RegularGridInterpolator` from S. Barrett:
    https://github.com/sbarratt/torch_interpolations
    published under Apache V2 license

    Copyright (C) 2020, S. Barrett

    Licensed to the Apache Software Foundation (ASF) under one
    or more contributor license agreements.  See the NOTICE file
    distributed with this work for additional information
    regarding copyright ownership.  The ASF licenses this file
    to you under the Apache License, Version 2.0 (the
    "License"); you may not use this file except in compliance
    with the License.  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing,
    software distributed under the License is distributed on an
    "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
    KIND, either express or implied.  See the License for the
    specific language governing permissions and limitations
    under the License.
    """

    def __init__(self, points, values):
        self.points = points
        self.values = values

        assert isinstance(self.points, tuple) or isinstance(self.points, list)
        assert isinstance(self.values, torch.Tensor)

        self.ms = list(self.values.shape)
        self.n = len(self.points)

        assert len(self.ms) == self.n

        for i, p in enumerate(self.points):
            assert isinstance(p, torch.Tensor)
            assert p.shape[0] == self.values.shape[i]

    def __call__(self, points_to_interp):
        assert self.points is not None
        assert self.values is not None

        assert len(points_to_interp) == len(self.points)
        K = points_to_interp[0].shape[0]
        for x in points_to_interp:
            assert x.shape[0] == K

        idxs = []
        dists = []
        overalls = []
        for p, x in zip(self.points, points_to_interp):
            idx_right = torch.bucketize(x, p)
            idx_right[idx_right >= p.shape[0]] = p.shape[0] - 1
            idx_left = (idx_right - 1).clamp(0, p.shape[0] - 1)
            dist_left = x - p[idx_left]
            dist_right = p[idx_right] - x
            dist_left[dist_left < 0] = 0.0
            dist_right[dist_right < 0] = 0.0
            both_zero = (dist_left == 0) & (dist_right == 0)
            dist_left[both_zero] = dist_right[both_zero] = 1.0

            idxs.append((idx_left, idx_right))
            dists.append((dist_left, dist_right))
            overalls.append(dist_left + dist_right)

        numerator = 0.0
        for indexer in product([0, 1], repeat=self.n):
            as_s = [idx[onoff] for onoff, idx in zip(indexer, idxs)]
            bs_s = [dist[1 - onoff] for onoff, dist in zip(indexer, dists)]
            numerator += self.values[as_s] * torch.prod(torch.stack(bs_s), dim=0)
        denominator = torch.prod(torch.stack(overalls), dim=0)
        return numerator / denominator


def interp1d(x_eval: torch.Tensor, x_dat: torch.Tensor, y_dat: torch.Tensor):
    """1D bilinear interpolation

    simple torch implementation of :func:`numpy.interp`

    Args:
        x_eval (torch.Tensor): The x-coordinates at which to evaluate the interpolated values.
        x_dat (torch.Tensor): The x-coordinates of the data points
        y_dat (torch.Tensor): The y-coordinates of the data points, same length as `x_dat`.

    Returns:
        torch.Tensor: The interpolated values, same shape as `x_eval`
    """
    assert len(x_dat) == len(y_dat)
    assert not torch.is_complex(x_dat)

    # sort x input data
    i_sort = torch.argsort(x_dat)
    _x = x_dat[i_sort]
    _y = y_dat[i_sort]

    # find left/right neighbor x datapoints
    idx_r = torch.bucketize(x_eval, _x)
    idx_l = idx_r - 1
    idx_r = idx_r.clamp(0, _x.shape[0] - 1)
    idx_l = idx_l.clamp(0, _x.shape[0] - 1)

    # distances to left / right (=weights)
    dist_l = x_eval - _x[idx_l]
    dist_r = _x[idx_r] - x_eval
    dist_l[dist_l < 0] = 0.0
    dist_r[dist_r < 0] = 0.0
    dist_l[torch.logical_and(dist_l == 0, dist_r == 0)] = 1.0
    sum_d_l_r = dist_l + dist_r
    y_l = _y[idx_l]
    y_r = _y[idx_r]

    # bilinear interpolated values
    y_eval = (y_l * dist_r + y_r * dist_l) / sum_d_l_r

    return y_eval



# encoding=utf-8
# %%
import unittest

import torch

import torchgdm as tg


class TestFreeSpacePlanWave(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing plane wave illumination.")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # --- setup a test case
        eps_env = 1.3
        inangledeg = 180.0
        e_p = 1
        e_s = 1
        self.r_probe = torch.as_tensor(
            [[100, 0, 20], [0, 0, 100], [-100, 0, 200], [-500, 40, -200]],
            dtype=tg.constants.DTYPE_FLOAT,
        )

        # --- setup a reference simulation
        self.env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=eps_env)

        self.efield = tg.env.freespace_3d.PlaneWave(
            e0p=e_p, e0s=e_s, inc_angle=inangledeg * torch.pi / 180
        )

    def test_E_H_factor(self):
        """test if in homogeneous medium |H|=n_env*|E|"""
        for device in self.devices:
            r_probe = (torch.rand((10, 3), device=device) - 0.5) * 1000
            wavelength = torch.tensor(500.0, device=device)
            for eps_env in [1.0, 2.25, 9.0]:
                efield = tg.env.freespace_3d.PlaneWave(device=device)
                env = tg.env.freespace_3d.EnvHomogeneous3D(
                    env_material=eps_env, device=device
                )
                E0tg = efield.get_efield(
                    r_probe, wavelength=wavelength, environment=env
                )
                H0tg = efield.get_hfield(
                    r_probe, wavelength=wavelength, environment=env
                )

                torch.testing.assert_close(
                    torch.sum((H0tg.abs()[:, 1]) / (E0tg.abs()[:, 0]) - eps_env**0.5),
                    torch.as_tensor(0.0, device=device),
                )

    def test_phase(self):
        config_pairs = [
            # test-case 1
            [
                dict(
                    e0s=1.0,
                    e0p=0.0,
                    inc_angle=0.0,
                    phase_e0s=torch.pi / 2,
                    inc_plane="xz",
                ),
                dict(
                    e0s=1.0,
                    e0p=0.0,
                    inc_angle=0.0,
                    phase_global=torch.pi / 2,
                    inc_plane="xz",
                ),
            ],
            # test-case 2
            [
                dict(e0s=1.0, e0p=1.0, inc_angle=torch.pi / 2, inc_plane="xz"),
                dict(
                    e0s=1.0,
                    e0p=1.0,
                    inc_angle=torch.pi / 2,
                    phase_global=torch.pi * 2,
                    inc_plane="xz",
                ),
            ],
            # test-case 3
            [
                dict(inc_plane="yz"),
                dict(phase_global=torch.pi * 2, inc_plane="yz"),
            ],
        ]
        for device in self.devices:
            wavelength = torch.as_tensor(500.0, device=device)
            for conf1, conf2 in config_pairs:
                env = tg.env.freespace_3d.EnvHomogeneous3D(device=device)
                ef1 = tg.env.freespace_3d.PlaneWave(**conf1, device=device)
                ef2 = tg.env.freespace_3d.PlaneWave(**conf2, device=device)

                r_probe = torch.as_tensor(
                    [[120, -50, -100]], dtype=tg.constants.DTYPE_FLOAT, device=device
                )

                e01 = ef1.get_efield(r_probe, wavelength=wavelength, environment=env)
                h01 = ef1.get_hfield(r_probe, wavelength=wavelength, environment=env)
                e02 = ef1.get_efield(r_probe, wavelength=wavelength, environment=env)
                h02 = ef1.get_hfield(r_probe, wavelength=wavelength, environment=env)

                torch.testing.assert_close(e01, e02)
                torch.testing.assert_close(h01, h02)

    def test_vector_directions(self):
        for device in self.devices:
            configs = [
                dict(e0s=0.0, e0p=1.0, inc_angle=0.0, inc_plane="xz"),
                dict(e0s=0.0, e0p=1.0, inc_angle=torch.pi / 2, inc_plane="xz"),
                dict(e0s=0.0, e0p=1.0, inc_angle=torch.pi / 2, inc_plane="yz"),
                dict(e0s=1.0, e0p=0.0, inc_angle=0.0, inc_plane="xz"),
                dict(e0s=1.0, e0p=0.0, inc_angle=torch.pi / 2, inc_plane="xz"),
                dict(e0s=1.0, e0p=0.0, inc_angle=torch.pi / 2, inc_plane="yz"),
                dict(
                    e0s=1.0, e0p=0.0, inc_angle=0.0, phase_e0s=torch.pi / 4, inc_plane="xz"
                ),
                dict(
                    e0s=1.0,
                    e0p=0.0,
                    inc_angle=torch.pi / 2,
                    phase_e0s=torch.pi / 4,
                    inc_plane="xz",
                ),
                dict(
                    e0s=1.0,
                    e0p=0.0,
                    inc_angle=torch.pi / 2,
                    phase_e0s=torch.pi / 4,
                    inc_plane="yz",
                ),
            ]
            wavelength = torch.as_tensor(500.0, device=device)
            for conf in configs:
                env = tg.env.freespace_3d.EnvHomogeneous3D(device=device)
                ef = tg.env.freespace_3d.PlaneWave(**conf, device=device)

                r_probe = torch.as_tensor(
                    [[0, 0, 0]], dtype=tg.constants.DTYPE_FLOAT, device=device
                )

                e0 = ef.get_efield(r_probe, wavelength=wavelength, environment=env)
                h0 = ef.get_hfield(r_probe, wavelength=wavelength, environment=env)
                k_unit = ef.k_vec_unit

                # check orthogonality
                torch.testing.assert_close(
                    torch.einsum("in,in->i", e0, h0).sum(),
                    torch.as_tensor(0 + 0j, device=device),
                )
                torch.testing.assert_close(
                    torch.einsum("in,in->i", e0, k_unit.unsqueeze(0)).sum(),
                    torch.as_tensor(0 + 0j, device=device),
                )

                # check handedness of vectors
                e_cross_h_dot_k = torch.einsum(
                    "in,in->i", torch.cross(e0, h0, dim=1), k_unit.unsqueeze(0)
                )
                if e_cross_h_dot_k[0].real != 0:
                    torch.testing.assert_close(
                        torch.sign(e_cross_h_dot_k[0].real),
                        torch.as_tensor(1.0, device=device),
                    )
                if e_cross_h_dot_k[0].imag != 0:
                    torch.testing.assert_close(
                        torch.sign(e_cross_h_dot_k[0].imag),
                        torch.as_tensor(1.0, device=device),
                    )

    def test_calculate_E_H(self):
        wavelength = 500.0
        for device in self.devices:
            self.r_probe = self.r_probe.to(device=device)
            self.efield.set_device(device)
            self.env.set_device(device)

            e0 = self.efield.get_efield(
                self.r_probe, wavelength=wavelength, environment=self.env
            )
            h0 = self.efield.get_hfield(
                self.r_probe, wavelength=wavelength, environment=self.env
            )
            # print(e0)
            # print(h0)

            e0_truth = torch.as_tensor(
                [
                    [
                        9.5922262e-01 - 2.8265165e-01j,
                        9.5922262e-01 - 2.8265165e-01j,
                        -1.1747089e-16 + 3.4614844e-17j,
                    ],
                    [
                        1.3757181e-01 - 9.9049180e-01j,
                        1.3757181e-01 - 9.9049180e-01j,
                        -1.6847687e-17 + 1.2130026e-16j,
                    ],
                    [
                        -9.6214800e-01 - 2.7252749e-01j,
                        -9.6214800e-01 - 2.7252749e-01j,
                        1.1782915e-16 + 3.3374992e-17j,
                    ],
                    [
                        -9.6214800e-01 + 2.7252749e-01j,
                        -9.6214800e-01 + 2.7252749e-01j,
                        1.1782915e-16 - 3.3374992e-17j,
                    ],
                ],
                dtype=tg.constants.DTYPE_COMPLEX,
                device=device,
            )

            h0_truth = torch.as_tensor(
                [
                    [
                        1.0936821e00 - 3.2227247e-01j,
                        -1.0936821e00 + 3.2227247e-01j,
                        -1.3393742e-16 + 3.9466995e-17j,
                    ],
                    [
                        1.5685599e-01 - 1.1293344e00j,
                        -1.5685599e-01 + 1.1293344e00j,
                        -1.9209319e-17 + 1.3830358e-16j,
                    ],
                    [
                        -1.0970175e00 - 3.1072915e-01j,
                        1.0970175e00 + 3.1072915e-01j,
                        1.3434590e-16 + 3.8053346e-17j,
                    ],
                    [
                        -1.0970175e00 + 3.1072915e-01j,
                        1.0970175e00 - 3.1072915e-01j,
                        1.3434590e-16 - 3.8053346e-17j,
                    ],
                ],
                dtype=tg.constants.DTYPE_COMPLEX,
                device=device,
            )

            # compare random e-field
            torch.testing.assert_close(e0, e0_truth)
            torch.testing.assert_close(h0, h0_truth)
            if self.verbose:
                print("  - {}: field evaluation test passed.".format(device))

    def test_integration_crosssections(self):

        for device in self.devices:
            torch.set_printoptions(precision=3, linewidth=200)

            # - anisotropic structure
            step = 40.0
            mat_struct = tg.materials.MatConstant(eps=20 + 0.5j)
            mat_struct = tg.materials.MatDatabase("Ge")
            struct = tg.struct.StructDiscretizedCubic3D(
                tg.struct.volume.cuboid(l=4, w=4, h=4), step, mat_struct
            )

            # - environment / illuminations setup
            env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=1)
            wavelengths = torch.as_tensor([550])

            # illumination configs for same electric response
            e_inc_same_e_1 = [  # Ex, Hy, kz   -   Ex, Hz, ky
                tg.env.freespace_3d.PlaneWave(
                    e0p=0, e0s=1, inc_angle=0, inc_plane="yz"
                ),
                tg.env.freespace_3d.PlaneWave(
                    e0p=0, e0s=1, inc_angle=torch.pi / 2, inc_plane="yz"
                ),
            ]
            e_inc_same_e_2 = [
                tg.env.freespace_3d.PlaneWave(
                    e0p=0, e0s=1, inc_angle=0, inc_plane="xz"
                ),
                tg.env.freespace_3d.PlaneWave(
                    e0p=0, e0s=1, inc_angle=torch.pi / 2, inc_plane="xz"
                ),
            ]
            e_inc_same_e_3 = [
                tg.env.freespace_3d.PlaneWave(
                    e0p=1, e0s=0, inc_angle=torch.pi / 2, inc_plane="xz"
                ),
                tg.env.freespace_3d.PlaneWave(
                    e0p=1, e0s=0, inc_angle=torch.pi / 2, inc_plane="yz"
                ),
            ]

            # illumination configs for same magnetic response
            e_inc_same_h_1 = [  # Ey, Hx, kz   -   Ez, Hx, ky
                tg.env.freespace_3d.PlaneWave(
                    e0p=1, e0s=0, inc_angle=0, inc_plane="yz"
                ),
                tg.env.freespace_3d.PlaneWave(
                    e0p=1, e0s=0, inc_angle=torch.pi / 2, inc_plane="yz"
                ),
            ]
            e_inc_same_h_2 = [
                tg.env.freespace_3d.PlaneWave(
                    e0p=1, e0s=0, inc_angle=0, inc_plane="xz"
                ),
                tg.env.freespace_3d.PlaneWave(
                    e0p=1, e0s=0, inc_angle=torch.pi / 2, inc_plane="xz"
                ),
            ]
            e_inc_same_h_3 = [
                tg.env.freespace_3d.PlaneWave(
                    e0p=0, e0s=1, inc_angle=torch.pi / 2, inc_plane="xz"
                ),
                tg.env.freespace_3d.PlaneWave(
                    e0p=0, e0s=1, inc_angle=torch.pi / 2, inc_plane="yz"
                ),
            ]

            # symmetric structure: test all illumination pairs
            e_inc_list = [
                e_inc_same_e_1,
                e_inc_same_e_2,
                e_inc_same_e_3,
                e_inc_same_h_1,
                e_inc_same_h_2,
                e_inc_same_h_3,
            ]

            # test all pairs
            for i, e_inc in enumerate(e_inc_list):
                sim = tg.simulation.Simulation(
                    structures=[struct],
                    environment=env,
                    illumination_fields=e_inc,
                    wavelengths=wavelengths,
                    device=device,
                )
                sim.run(verbose=False, progress_bar=False)

                # - scs
                scs_mp = sim.get_spectra_multipole_scs(progress_bar=False)
                torch.testing.assert_close(
                    scs_mp["scs_ed"][:, 0], scs_mp["scs_ed"][:, 1], atol=0.5, rtol=1e-3
                )
                torch.testing.assert_close(
                    scs_mp["scs_eq"][:, 0], scs_mp["scs_eq"][:, 1], atol=0.5, rtol=1e-3
                )
                torch.testing.assert_close(
                    scs_mp["scs_md"][:, 0], scs_mp["scs_md"][:, 1], atol=0.5, rtol=1e-3
                )
                torch.testing.assert_close(
                    scs_mp["scs_mq"][:, 0], scs_mp["scs_mq"][:, 1], atol=0.5, rtol=1e-3
                )

                # - ecs
                ecs_mp = sim.get_spectra_multipole_ecs(progress_bar=False)
                torch.testing.assert_close(
                    ecs_mp["ecs_ed"][:, 0], ecs_mp["ecs_ed"][:, 1], atol=0.5, rtol=1e-3
                )
                torch.testing.assert_close(
                    ecs_mp["ecs_eq"][:, 0], ecs_mp["ecs_eq"][:, 1], atol=0.5, rtol=1e-3
                )
                torch.testing.assert_close(
                    ecs_mp["ecs_md"][:, 0], ecs_mp["ecs_md"][:, 1], atol=0.5, rtol=1e-3
                )
                torch.testing.assert_close(
                    ecs_mp["ecs_mq"][:, 0], ecs_mp["ecs_mq"][:, 1], atol=0.5, rtol=1e-3
                )

                # print("------")
                # print(i, "scs:")
                # print("ED - ", scs_mp["scs_ed"][:, 0], scs_mp["scs_ed"][:, 1])
                # print("EQ - ", scs_mp["scs_eq"][:, 0], scs_mp["scs_eq"][:, 1])
                # print("MD - ", scs_mp["scs_md"][:, 0], scs_mp["scs_md"][:, 1])
                # print("MQ - ", scs_mp["scs_mq"][:, 0], scs_mp["scs_mq"][:, 1])
                # print(i, "ecs:")
                # print("ED - ", ecs_mp["ecs_ed"][:, 0], ecs_mp["ecs_ed"][:, 1])
                # print("EQ - ", ecs_mp["ecs_eq"][:, 0], ecs_mp["ecs_eq"][:, 1])
                # print("MD - ", ecs_mp["ecs_md"][:, 0], ecs_mp["ecs_md"][:, 1])
                # print("MQ - ", ecs_mp["ecs_mq"][:, 0], ecs_mp["ecs_mq"][:, 1])


class TestFreeSpaceGaussian(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing gaussian beam illumination.")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # --- setup a test case
        eps_env = 1.45**2
        inangledeg = 40.0 * torch.pi / 180.0
        e_p = 1
        e_s = 0.5
        waist = 250.0
        self.r_focus = torch.as_tensor([0, 0, 10])
        self.r_probe = torch.as_tensor(
            [[100, 0, 20], [0, 0, 100], [-100, 0, 200], [-500, 40, -200]],
            dtype=tg.constants.DTYPE_FLOAT,
        )

        # --- setup a reference simulation
        self.env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=eps_env)

        self.efield = tg.env.freespace_3d.GaussianParaxial(
            e0p=e_p,
            e0s=e_s,
            r_focus=self.r_focus,
            waist=waist,
            inc_angle=inangledeg,
            correction_div_e=True,
        )

    def test_compare_with_planewave(self):
        wavelength = 650.0
        for device in self.devices:
            self.r_probe = self.r_probe.to(device=device)
            self.env.set_device(device)
            polas = [[1, 0], [0, 1], [1, 1]]
            inc_angles_deg = [0, 30, 90, 150, 180]
            for e_p_s in polas:
                for inc_angle_deg in inc_angles_deg:

                    planewave = tg.env.freespace_3d.PlaneWave(
                        e0p=e_p_s[0],
                        e0s=e_p_s[1],
                        inc_angle=inc_angle_deg * torch.pi / 180,
                        device=device,
                    )

                    gaussian = tg.env.freespace_3d.GaussianParaxial(
                        e0p=e_p_s[0],
                        e0s=e_p_s[1],
                        r_focus=[0, 0, 0],
                        waist=10000000,
                        inc_angle=inc_angle_deg * torch.pi / 180,
                        correction_div_e=True,
                        device=device,
                    )

                    # eval electric fields
                    e0_p = planewave.get_efield(
                        [0, 0, 0], wavelength=wavelength, environment=self.env
                    )
                    e0_g = gaussian.get_efield(
                        [0, 0, 0], wavelength=wavelength, environment=self.env
                    )
                    torch.testing.assert_close(e0_p, e0_g, rtol=1e-2, atol=0.01)

                    # eval magnetic fields
                    h0_p = planewave.get_hfield(
                        [0, 0, 0], wavelength=wavelength, environment=self.env
                    )
                    h0_g = gaussian.get_hfield(
                        [0, 0, 0], wavelength=wavelength, environment=self.env
                    )
                    torch.testing.assert_close(h0_p, h0_g, rtol=1e-2, atol=0.01)

    def test_calculate_E_H(self):
        wavelength = 650.0
        for device in self.devices:
            self.r_probe = self.r_probe.to(device=device)
            self.efield.set_device(device)
            self.env.set_device(device)

            e0 = self.efield.get_efield(
                self.r_probe, wavelength=wavelength, environment=self.env
            )
            h0 = self.efield.get_hfield(
                self.r_probe, wavelength=wavelength, environment=self.env
            )
            # print(e0)
            # print(h0)

            e0_truth = torch.as_tensor(
                [
                    [
                        -0.5448776 - 4.3501070e-01j,
                        0.3144340 + 3.2393885e-01j,
                        -0.3289866 - 4.8949170e-01j,
                    ],
                    [
                        -0.4636163 - 5.4261190e-01j,
                        0.3361847 + 3.1875351e-01j,
                        -0.4935032 - 3.4512645e-01j,
                    ],
                    [
                        -0.0536437 - 4.3212891e-01j,
                        0.1200230 + 2.3944278e-01j,
                        -0.3095120 - 2.3002407e-01j,
                    ],
                    [
                        -0.3082355 + 4.6097115e-04j,
                        0.1981485 - 2.4692476e-02j,
                        -0.2491881 + 7.6279074e-02j,
                    ],
                ],
                dtype=tg.constants.DTYPE_COMPLEX,
                device=device,
            )

            h0_truth = torch.as_tensor(
                [
                    [
                        -0.3950363 - 3.1538275e-01j,
                        -0.9118586 - 9.3942261e-01j,
                        -0.2385153 - 3.5488147e-01j,
                    ],
                    [
                        -0.3361218 - 3.9339361e-01j,
                        -0.9749354 - 9.2438513e-01j,
                        -0.3577898 - 2.5021666e-01j,
                    ],
                    [
                        -0.0388917 - 3.1329343e-01j,
                        -0.3480666 - 6.9438404e-01j,
                        -0.2243962 - 1.6676745e-01j,
                    ],
                    [
                        -0.2234707 + 3.3420406e-04j,
                        -0.5746307 + 7.1608178e-02j,
                        -0.1806614 + 5.5302326e-02j,
                    ],
                ],
                dtype=tg.constants.DTYPE_COMPLEX,
                device=device,
            )

            # compare random e-field
            torch.testing.assert_close(e0, e0_truth)
            torch.testing.assert_close(h0, h0_truth)
            if self.verbose:
                print("  - {}: field evaluation test passed.".format(device))


class TestFreeSpaceElectricDipole(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing electric dipole illumination.")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # --- setup a test case
        eps_env = 1.3
        p_source = torch.as_tensor([1.0, -1.0, 0.5], dtype=tg.constants.DTYPE_FLOAT)
        r_source = torch.as_tensor([10.0, 5.0, 15.0], dtype=tg.constants.DTYPE_FLOAT)
        self.r_probe = torch.as_tensor(
            [[10, 0, 20], [0, 0, -10], [-10, 10, 5], [-500, 40, -200]],
            dtype=tg.constants.DTYPE_FLOAT,
        )

        # --- setup a reference simulation
        self.env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=eps_env)

        self.efield = tg.env.freespace_3d.ElectricDipole(
            r_source=r_source, p_source=p_source
        )

    def test_calculate_E_H(self):
        wavelength = 500.0
        for device in self.devices:
            self.r_probe = self.r_probe.to(device=device)
            self.efield.set_device(device)
            self.env.set_device(device)

            e0 = self.efield.get_efield(
                self.r_probe, wavelength=wavelength, environment=self.env
            )
            h0 = self.efield.get_hfield(
                self.r_probe, wavelength=wavelength, environment=self.env
            )
            # print(e0)
            # print(h0)

            e0_truth = torch.as_tensor(
                [
                    [
                        -2.1646330e-03 + 1.5052938e-06j,
                        -2.7391179e-03 - 1.5064361e-06j,
                        3.8214340e-03 + 7.5378472e-07j,
                    ],
                    [
                        -7.9824858e-06 + 1.4676721e-06j,
                        4.8353097e-05 - 1.4596324e-06j,
                        4.9836235e-05 + 7.4455437e-07j,
                    ],
                    [
                        1.6251166e-04 + 1.4944962e-06j,
                        4.9544378e-06 - 1.4806619e-06j,
                        8.1255821e-05 + 7.4724721e-07j,
                    ],
                    [
                        8.1364007e-08 - 5.7865215e-09j,
                        5.1592448e-08 - 2.5644798e-07j,
                        2.9485673e-08 + 1.9189642e-08j,
                    ],
                ],
                dtype=tg.constants.DTYPE_COMPLEX,
                device=device,
            )

            h0_truth = torch.as_tensor(
                [
                    [
                        -3.0768206e-08 + 8.9312525e-05j,
                        -6.1536412e-08 + 1.7862505e-04j,
                        -6.1536412e-08 + 1.7862505e-04j,
                    ],
                    [
                        3.3363011e-07 - 1.8070623e-05j,
                        2.4264011e-07 - 1.3142271e-05j,
                        -1.8198000e-07 + 9.8567043e-06j,
                    ],
                    [
                        9.1412744e-08 - 8.2457755e-06j,
                        0.0000000e00 + 0.0000000e00j,
                        -1.8282549e-07 + 1.6491551e-05j,
                    ],
                    [
                        2.5095151e-08 - 1.1380054e-07j,
                        -5.0825637e-09 + 2.3048244e-08j,
                        -6.0355426e-08 + 2.7369757e-07j,
                    ],
                ],
                dtype=tg.constants.DTYPE_COMPLEX,
                device=device,
            )

            # compare random e-field
            torch.testing.assert_close(e0, e0_truth)
            torch.testing.assert_close(h0, h0_truth)
            if self.verbose:
                print("  - {}: field evaluation test passed.".format(device))


class TestFreeSpaceMagneticDipole(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing magnetic dipole illumination.")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # --- setup a test case
        eps_env = 1.3
        m_source = torch.as_tensor([1.0, -1.0, 0.5], dtype=tg.constants.DTYPE_FLOAT)
        r_source = torch.as_tensor([10.0, 5.0, 15.0], dtype=tg.constants.DTYPE_FLOAT)
        self.r_probe = torch.as_tensor(
            [[10, 0, 20], [0, 0, -10], [-10, 10, 5], [-500, 40, -200]],
            dtype=tg.constants.DTYPE_FLOAT,
        )

        # --- setup a reference simulation
        self.env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=eps_env)

        self.efield = tg.env.freespace_3d.MagneticDipole(
            r_source=r_source, m_source=m_source
        )

    def test_calculate_E_H(self):
        wavelength = 500.0
        for device in self.devices:
            self.r_probe = self.r_probe.to(device=device)
            self.efield.set_device(device)
            self.env.set_device(device)

            e0 = self.efield.get_efield(
                self.r_probe, wavelength=wavelength, environment=self.env
            )
            h0 = self.efield.get_hfield(
                self.r_probe, wavelength=wavelength, environment=self.env
            )
            # print(e0)
            # print(h0)

            e0_truth = torch.as_tensor(
                [
                    [
                        3.0768206e-08 - 8.9312525e-05j,
                        6.1536412e-08 - 1.7862505e-04j,
                        6.1536412e-08 - 1.7862505e-04j,
                    ],
                    [
                        -3.3363062e-07 + 1.8070623e-05j,
                        -2.4264045e-07 + 1.3142271e-05j,
                        1.8198034e-07 - 9.8567043e-06j,
                    ],
                    [
                        -9.1413085e-08 + 8.2457755e-06j,
                        0.0000000e00 + 0.0000000e00j,
                        1.8282617e-07 - 1.6491551e-05j,
                    ],
                    [
                        -2.5095151e-08 + 1.1380054e-07j,
                        5.0825637e-09 - 2.3048230e-08j,
                        6.0355426e-08 - 2.7369754e-07j,
                    ],
                ],
                dtype=tg.constants.DTYPE_COMPLEX,
                device=device,
            )

            h0_truth = torch.as_tensor(
                [
                    [
                        -2.8140228e-03 + 1.9568818e-06j,
                        -3.5608532e-03 - 1.9583667e-06j,
                        4.9678641e-03 + 9.7992017e-07j,
                    ],
                    [
                        -1.0377235e-05 + 1.9079737e-06j,
                        6.2859021e-05 - 1.8975221e-06j,
                        6.4787106e-05 + 9.6792064e-07j,
                    ],
                    [
                        2.1126511e-04 + 1.9428496e-06j,
                        6.4407595e-06 - 1.9248594e-06j,
                        1.0563256e-04 + 9.7142367e-07j,
                    ],
                    [
                        1.0577320e-07 - 7.5224733e-09j,
                        6.7070182e-08 - 3.3338233e-07j,
                        3.8331365e-08 + 2.4946530e-08j,
                    ],
                ],
                dtype=tg.constants.DTYPE_COMPLEX,
                device=device,
            )

            # compare random e-field
            torch.testing.assert_close(e0, e0_truth)
            torch.testing.assert_close(h0, h0_truth)
            if self.verbose:
                print("  - {}: field evaluation test passed.".format(device))


# %%


if __name__ == "__main__":
    print("testing illumination fields.")
    torch.set_printoptions(precision=7)
    unittest.main(argv=["first-arg-is-ignored"], exit=False)
